# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

import torch.nn as nn

from airtk.attack.base import Attack, LabelMixin
from airtk.util import (
    batch_multiply,
    clamp,
    normalize_by_pnorm,
)


class FGSMAttack(Attack, LabelMixin):
    """

    One step fast gradient sign method (Goodfellow et al, 2014).

    Arguments:

        predict (nn.Module): forward pass function.

        loss_fn (nn.Module): loss function.

        eps (float): attack step size.

        clip_min (float): mininum value per input dimension.

        clip_max (float): maximum value per input dimension.

        targeted (bool): indicate if this is a targeted attack.

    """

    def __init__(
        self,
        predict,
        loss_fn=None,
        eps=0.3,
        clip_min=0.0,
        clip_max=1.0,
        targeted=False,
    ):
        super(FGSMAttack, self).__init__(predict, loss_fn, clip_min, clip_max)

        self.eps = eps
        self.targeted = targeted

        if self.loss_fn is None:
            self.loss_fn = nn.CrossEntropyLoss(reduction="sum")

    def perturb(self, x, y=None):
        """

        Given examples (x, y), returns their adversarial counterparts with an attack length of eps.

        Arguments:

            x (torch.Tensor): input tensor.

            y  (torch.Tensor): label tensor.

                - if None and self.targeted=False, compute y as predicted labels.

                - if self.targeted=True, then y must be the targeted labels.

        Returns:

            torch.Tensor containing perturbed inputs.

            torch.Tensor containing the perturbation.

        """

        x, y = self._verify_and_process_inputs(x, y)

        xadv = x.requires_grad_()

        outputs = self.predict(xadv)

        loss = self.loss_fn(outputs, y)

        if self.targeted:
            loss = -loss

        loss.backward()

        grad_sign = xadv.grad.detach().sign()

        xadv = xadv + batch_multiply(self.eps, grad_sign)

        xadv = clamp(xadv, self.clip_min, self.clip_max)

        radv = xadv - x

        return xadv.detach(), radv.detach()


LinfFastGradientAttack = FGSMAttack


class FGMAttack(Attack, LabelMixin):
    """

    One step fast gradient method. Perturbs the input with gradient (not gradient sign) of the loss wrt the input.

    Arguments:

        predict (nn.Module): forward pass function.

        loss_fn (nn.Module): loss function.

        eps (float): attack step size.

        clip_min (float): mininum value per input dimension.

        clip_max (float): maximum value per input dimension.

        targeted (bool): indicate if this is a targeted attack.

    """

    def __init__(
        self,
        predict,
        loss_fn=None,
        eps=0.3,
        clip_min=0.0,
        clip_max=1.0,
        targeted=False,
    ):
        super(FGMAttack, self).__init__(predict, loss_fn, clip_min, clip_max)

        self.eps = eps
        self.targeted = targeted

        if self.loss_fn is None:
            self.loss_fn = nn.CrossEntropyLoss(reduction="sum")

    def perturb(self, x, y=None):
        """

        Given examples (x, y), returns their adversarial counterparts with an attack length of eps.

        Arguments:

            x (torch.Tensor): input tensor.

            y  (torch.Tensor): label tensor.

                - if None and self.targeted=False, compute y as predicted labels.

                - if self.targeted=True, then y must be the targeted labels.

        Returns:

            torch.Tensor containing perturbed inputs.

            torch.Tensor containing the perturbation.

        """

        x, y = self._verify_and_process_inputs(x, y)

        xadv = x.requires_grad_()

        outputs = self.predict(xadv)

        loss = self.loss_fn(outputs, y)

        if self.targeted:
            loss = -loss

        loss.backward()

        grad = normalize_by_pnorm(xadv.grad)

        xadv = xadv + batch_multiply(self.eps, grad)

        xadv = clamp(xadv, self.clip_min, self.clip_max)

        radv = xadv - x

        return xadv.detach(), radv.detach()


L2FastGradientAttack = FGMAttack



import torch
import torch.nn as nn

from ..attack import Attack


class SINIFGSM(Attack):
    r"""
    SI-NI-FGSM in the paper 'NESTEROV ACCELERATED GRADIENT AND SCALEINVARIANCE FOR ADVERSARIAL ATTACKS'
    [https://arxiv.org/abs/1908.06281], Published as a conference paper at ICLR 2020
    Modified from "https://githuba.com/JHL-HUST/SI-NI-FGSM"

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)
        alpha (float): step size. (Default: 2/255)
        steps (int): number of iterations. (Default: 10)
        decay (float): momentum factor. (Default: 1.0)
        m (int): number of scale copies. (Default: 5)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.SINIFGSM(model, eps=8/255, alpha=2/255, steps=10, decay=1.0, m=5)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self, model, eps=8 / 255, alpha=2 / 255, steps=10, decay=1.0, m=5
    ):
        super().__init__("SINIFGSM", model)
        self.eps = eps
        self.steps = steps
        self.decay = decay
        self.alpha = alpha
        self.m = m
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        momentum = torch.zeros_like(images).detach().to(self.device)

        loss = nn.CrossEntropyLoss()

        adv_images = images.clone().detach()

        for _ in range(self.steps):
            adv_images.requires_grad = True
            nes_image = adv_images + self.decay * self.alpha * momentum
            # Calculate sum the gradients over the scale copies of the input image
            adv_grad = torch.zeros_like(images).detach().to(self.device)
            for i in torch.arange(self.m):
                nes_images = nes_image / torch.pow(2, i)
                outputs = self.get_logits(nes_images)
                # Calculate loss
                if self.targeted:
                    cost = -loss(outputs, target_labels)
                else:
                    cost = loss(outputs, labels)
                adv_grad += torch.autograd.grad(
                    cost, adv_images, retain_graph=False, create_graph=False
                )[0]
            adv_grad = adv_grad / self.m

            # Update adversarial images
            grad = self.decay * momentum + adv_grad / torch.mean(
                torch.abs(adv_grad), dim=(1, 2, 3), keepdim=True
            )
            momentum = grad
            adv_images = adv_images.detach() + self.alpha * grad.sign()
            delta = torch.clamp(
                adv_images - images, min=-self.eps, max=self.eps
            )
            adv_images = torch.clamp(images + delta, min=0, max=1).detach()

        return adv_images



import torch
import torch.nn as nn
from torch.distributions.normal import Normal

# ATES (https://github.com/chawins/ates-minimal/blob/master/lib/mnist_model.py)
class MNIST_ATES(nn.Module):
    def __init__(self, num_classes=10, quantize=False):
        super().__init__()
        self.quantize = quantize
        self.conv1 = nn.Conv2d(1, 64, kernel_size=8, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(64, 128, kernel_size=6, stride=2, padding=3)
        self.relu2 = nn.ReLU(inplace=True)
        self.conv3 = nn.Conv2d(128, 128, kernel_size=5, stride=1, padding=0)
        self.relu3 = nn.ReLU(inplace=True)
        self.fc = nn.Linear(2048, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        if self.quantize:
            x = torch.quantize_per_tensor(x, 1 / 16, 0, torch.qint8).dequantize()
        x = self.conv1(x)
        x = self.relu1(x)
        x = self.conv2(x)
        x = self.relu2(x)
        x = self.conv3(x)
        x = self.relu3(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        return x



def get_subloader(given_loader, n_limit):
    if n_limit is None:
        return given_loader

    sub_loader = []
    num = 0
    for item in given_loader:
        sub_loader.append(item)
        if isinstance(item, tuple) or isinstance(item, list):
            batch_size = len(item[0])
        else:
            batch_size = len(item)
        num += batch_size
        if num >= n_limit:
            break
    return sub_loader



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

import numpy as np
import torch
import torch.nn as nn

from airtk.attack.base import Attack, LabelMixin
from airtk.util import (
    batch_clamp,
    batch_multiply,
    clamp,
    clamp_by_pnorm,
    is_float_or_torch_tensor,
    normalize_by_pnorm,
    rand_init_delta,
)


def perturb_iterative(
    xvar,
    yvar,
    predict,
    nb_iter,
    eps,
    eps_iter,
    loss_fn,
    delta_init=None,
    minimize=False,
    ord=np.inf,
    clip_min=0.0,
    clip_max=1.0,
):
    """

    Iteratively maximize the loss over the input. It is a shared method for iterative atts.

    Arguments:

        xvar (torch.Tensor): input data.

        yvar (torch.Tensor): input labels.

        predict (nn.Module): forward pass function.

        nb_iter (int): number of iterations.

        eps (float): maximum distortion.

        eps_iter (float): attack step size.

        loss_fn (nn.Module): loss function.

        delta_init (torch.Tensor): (optional) tensor contains the random initialization.

        minimize (bool): (optional) whether to minimize or maximize the loss.

        ord (int): (optional) the order of maximum distortion (inf or 2).

        clip_min (float): mininum value per input dimension.

        clip_max (float): maximum value per input dimension.

    Returns:

        torch.Tensor containing the perturbed input,

        torch.Tensor containing the perturbation

    """

    if delta_init is not None:
        delta = delta_init

    else:
        delta = torch.zeros_like(xvar)

    delta.requires_grad_()

    for _ in range(nb_iter):
        outputs = predict(xvar + delta)

        loss = loss_fn(outputs, yvar)

        if minimize:
            loss = -loss

        loss.backward()

        if ord == np.inf:
            grad_sign = delta.grad.data.sign()

            delta.data = delta.data + batch_multiply(eps_iter, grad_sign)

            delta.data = batch_clamp(eps, delta.data)
            delta.data = (
                clamp(xvar.data + delta.data, clip_min, clip_max) - xvar.data
            )

        elif ord == 2:
            grad = delta.grad.data

            grad = normalize_by_pnorm(grad)

            delta.data = delta.data + batch_multiply(eps_iter, grad)
            delta.data = (
                clamp(xvar.data + delta.data, clip_min, clip_max) - xvar.data
            )

            if eps is not None:
                delta.data = clamp_by_pnorm(delta.data, ord, eps)

        else:
            error = "Only ord=inf and ord=2 have been implemented"

            raise NotImplementedError(error)

        delta.grad.data.zero_()

    x_adv = clamp(xvar + delta, clip_min, clip_max)

    r_adv = x_adv - xvar

    return x_adv, r_adv


class PGDAttack(Attack, LabelMixin):
    """

    The projected gradient descent attack (Madry et al, 2017).

    The attack performs nb_iter steps of size eps_iter, while always staying within eps from the initial point.

    Arguments:

        predict (nn.Module): forward pass function.

        loss_fn (nn.Module): loss function.

        eps (float): maximum distortion.

        nb_iter (int): number of iterations.

        eps_iter (float): attack step size.

        rand_init (bool): (optional) random initialization.

        clip_min (float): mininum value per input dimension.

        clip_max (float): maximum value per input dimension.

        ord (int): (optional) the order of maximum distortion (inf or 2).

        targeted (bool): if the attack is targeted.

        rand_init_type (str): (optional) random initialization type.

    """

    def __init__(
        self,
        predict,
        loss_fn=None,
        eps=0.3,
        nb_iter=40,
        eps_iter=0.01,
        rand_init=True,
        clip_min=0.0,
        clip_max=1.0,
        ord=np.inf,
        targeted=False,
        rand_init_type="uniform",
    ):
        super().__init__(predict, loss_fn, clip_min, clip_max)
        self.eps = eps

        self.nb_iter = nb_iter
        self.eps_iter = eps_iter
        self.rand_init = rand_init

        self.rand_init_type = rand_init_type
        self.ord = ord

        self.targeted = targeted

        if self.loss_fn is None:
            self.loss_fn = nn.CrossEntropyLoss(reduction="sum")

        assert is_float_or_torch_tensor(self.eps_iter)

        assert is_float_or_torch_tensor(self.eps)

    def perturb(self, x, y=None):
        """

        Given examples (x, y), returns their adversarial counterparts with an attack length of eps.

        Arguments:

            x (torch.Tensor): input tensor.

            y (torch.Tensor): label tensor.

                - if None and self.targeted=False, compute y as predicted

                labels.

                - if self.targeted=True, then y must be the targeted labels.

        Returns:

            torch.Tensor containing perturbed inputs,

            torch.Tensor containing the perturbation

        """

        x, y = self._verify_and_process_inputs(x, y)

        delta = torch.zeros_like(x)

        delta = nn.Parameter(delta)

        if self.rand_init:
            if self.rand_init_type == "uniform":
                rand_init_delta(
                    delta, x, self.ord, self.eps, self.clip_min, self.clip_max
                )
                delta.data = (
                    clamp(x + delta.data, min=self.clip_min, max=self.clip_max)
                    - x
                )

            elif self.rand_init_type == "normal":
                delta.data = 0.001 * torch.randn_like(
                    x
                )  # initialize as in TRADES

            else:
                raise NotImplementedError(
                    "Only rand_init_type=normal and rand_init_type=uniform have been implemented."
                )

        x_adv, r_adv = perturb_iterative(
            x,
            y,
            self.predict,
            nb_iter=self.nb_iter,
            eps=self.eps,
            eps_iter=self.eps_iter,
            loss_fn=self.loss_fn,
            minimize=self.targeted,
            ord=self.ord,
            clip_min=self.clip_min,
            clip_max=self.clip_max,
            delta_init=delta,
        )

        return x_adv.data, r_adv.data


class LinfPGDAttack(PGDAttack):
    """

    PGD Attack with order=Linf

    Arguments:

        predict (nn.Module): forward pass function.

        loss_fn (nn.Module): loss function.

        eps (float): maximum distortion.

        nb_iter (int): number of iterations.

        eps_iter (float): attack step size.

        rand_init (bool): (optional) random initialization.

        clip_min (float): mininum value per input dimension.

        clip_max (float): maximum value per input dimension.

        targeted (bool): if the attack is targeted.

        rand_init_type (str): (optional) random initialization type.

    """

    def __init__(
        self,
        predict,
        loss_fn=None,
        eps=0.3,
        nb_iter=40,
        eps_iter=0.01,
        rand_init=True,
        clip_min=0.0,
        clip_max=1.0,
        targeted=False,
        rand_init_type="uniform",
    ):
        ord = np.inf

        super(LinfPGDAttack, self).__init__(
            predict=predict,
            loss_fn=loss_fn,
            eps=eps,
            nb_iter=nb_iter,
            eps_iter=eps_iter,
            rand_init=rand_init,
            clip_min=clip_min,
            clip_max=clip_max,
            targeted=targeted,
            ord=ord,
            rand_init_type=rand_init_type,
        )


class L2PGDAttack(PGDAttack):
    """

    PGD Attack with order=L2

    Arguments:

        predict (nn.Module): forward pass function.

        loss_fn (nn.Module): loss function.

        eps (float): maximum distortion.

        nb_iter (int): number of iterations.

        eps_iter (float): attack step size.

        rand_init (bool): (optional) random initialization.

        clip_min (float): mininum value per input dimension.

        clip_max (float): maximum value per input dimension.

        targeted (bool): if the attack is targeted.

        rand_init_type (str): (optional) random initialization type.

    """

    def __init__(
        self,
        predict,
        loss_fn=None,
        eps=0.3,
        nb_iter=40,
        eps_iter=0.01,
        rand_init=True,
        clip_min=0.0,
        clip_max=1.0,
        targeted=False,
        rand_init_type="uniform",
    ):
        ord = 2

        super(L2PGDAttack, self).__init__(
            predict=predict,
            loss_fn=loss_fn,
            eps=eps,
            nb_iter=nb_iter,
            eps_iter=eps_iter,
            rand_init=rand_init,
            clip_min=clip_min,
            clip_max=clip_max,
            targeted=targeted,
            ord=ord,
            rand_init_type=rand_init_type,
        )



import torch
import torch.nn as nn
import torch.nn.functional as F

from ...attacks import TPGD

from .advtrainer import AdvTrainer


class TRADES(AdvTrainer):
    r"""
    Adversarial training in 'Towards Deep Learning Models Resistant to Adversarial Attacks'
    [https://arxiv.org/abs/1706.06083]

    Attributes:
        self.rmodel : rmodel.
        self.device : device where rmodel is.
        self.optimizer : optimizer.
        self.scheduler : scheduler (Automatically updated).
        self.curr_epoch : current epoch starts from 1 (Automatically updated).
        self.curr_iter : current iters starts from 1 (Automatically updated).

    Arguments:
        rmodel (nn.Module): rmodel to train.
        eps (float): strength of the attack or maximum perturbation.
        alpha (float): step size.
        steps (int): number of steps.
        beta (float): trade-off regularization parameter.
    """

    def __init__(self, rmodel, eps, alpha, steps, beta):
        super().__init__(rmodel)
        self.atk = TPGD(rmodel, eps, alpha, steps)
        self.beta = beta

    def calculate_cost(self, train_data, reduction="mean"):
        r"""
        Overridden.
        """
        images, labels = train_data
        images = images.to(self.device)
        labels = labels.to(self.device)

        logits_clean = self.rmodel(images)
        loss_ce = nn.CrossEntropyLoss(reduction=reduction)(logits_clean, labels)

        adv_images = self.atk(images)
        logits_adv = self.rmodel(adv_images)
        probs_clean = F.softmax(logits_clean, dim=1)
        log_probs_adv = F.log_softmax(logits_adv, dim=1)
        loss_kl = nn.KLDivLoss(reduction="none")(log_probs_adv, probs_clean).sum(dim=1)

        cost = loss_ce + self.beta * loss_kl

        self.add_record_item("Loss", cost.mean().item())
        self.add_record_item("CELoss", loss_ce.mean().item())
        self.add_record_item("KLLoss", loss_kl.mean().item())

        return cost.mean() if reduction == "mean" else cost



import torch
import torch.nn.functional as F
import numpy as np

from ..attack import Attack


class PIFGSM(Attack):
    r"""
    PIFGSM in the paper 'Patch-wise Attack for Fooling Deep Neural Network'
    [https://arxiv.org/abs/2007.06765]

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        max_epsilon (float): maximum size of adversarial perturbation. (Default: 16/255)
        num_iter_set (float): number of iterations. (Default: 10)
        momentum (float): momentum. (Default: 1.0)
        amplification (float): to amplifythe step size. (Default: 10.0)
        prob (float): probability of using diverse inputs. (Default: 0.7)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.PIFGSM(model, eps=16/255, num_iter_set=10)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        max_epsilon=16 / 255,
        num_iter_set=10,
        momentum=1.0,
        amplification=10.0,
        prob=0.7,
    ):
        super().__init__("PIFGSM", model)
        self.max_epsilon = max_epsilon
        self.num_iter_set = num_iter_set
        self.momentum = momentum
        self.amplification = amplification
        self.prob = prob
        self.supported_mode = ["default"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        images_min = self.clip_by_tensor(images - self.max_epsilon, t_min=0, t_max=1)
        images_max = self.clip_by_tensor(images + self.max_epsilon, t_min=0, t_max=1)
        adv_images = self.graph(images, labels, images_min, images_max)

        return adv_images

    def clip_by_tensor(self, t, t_min, t_max):
        result = (t >= t_min).float() * t + (t < t_min).float() * t_min
        result = (result <= t_max).float() * result + (result > t_max).float() * t_max
        return result

    def project_noise(self, images, P_kern, padding_size):
        images = F.conv2d(
            images, P_kern, padding=(padding_size, padding_size), groups=3
        )
        return images

    def project_kern(self, kern_size):
        kern = np.ones((kern_size, kern_size), dtype=np.float32) / (kern_size ** 2 - 1)
        kern[kern_size // 2, kern_size // 2] = 0.0
        kern = kern.astype(np.float32)
        stack_kern = np.stack([kern, kern, kern])
        stack_kern = np.expand_dims(stack_kern, 1)
        stack_kern = torch.tensor(stack_kern).to(self.device)
        return stack_kern, kern_size // 2

    def graph(self, images, labels, images_min, images_max):
        eps = self.max_epsilon
        num_iter = self.num_iter_set
        alpha = eps / num_iter
        alpha_beta = alpha * self.amplification
        gamma = alpha_beta
        P_kern, padding_size = self.project_kern(3)

        images.requires_grad = True
        amplification = 0.0
        for _ in range(num_iter):
            # zero_gradients(images)
            if images.grad is not None:
                images.grad.detach_()
                images.grad.zero_()

            output_v3 = self.get_logits(images)
            loss = F.cross_entropy(output_v3, labels)
            loss.backward()
            noise = images.grad.data

            amplification += alpha_beta * torch.sign(noise)
            cut_noise = torch.clamp(abs(amplification) - eps, 0, 10000.0) * torch.sign(
                amplification
            )
            projection = gamma * torch.sign(
                self.project_noise(cut_noise, P_kern, padding_size)
            )
            amplification += projection

            images = images + alpha_beta * torch.sign(noise) + projection
            images = self.clip_by_tensor(images, images_min, images_max)
            images = images.detach().requires_grad_(True)

        return images.detach()



from collections import defaultdict

import torch


class Minimizer(object):
    def __init__(self, model, optimizer):
        self.model = model
        self.optimizer = optimizer
        self.state = defaultdict(dict)
        self.records = {}

    def update_records(self, key, value):
        arr = self.records.get(key)
        if arr is None:
            self.records[key] = []
            arr = self.records.get(key)
        arr.append(value)

    def state_dict(self):
        """Returns the state of the minimizer as a :class:`dict`.
        It contains an entry for every variable in self.__dict__ which
        is not the optimizer.
        """
        return {
            key: value
            for key, value in self.__dict__.items()
            if key not in ["model", "optimizer", "state"]
        }

    def load_state_dict(self, state_dict):
        """Loads the minimizers state.
        Args:
            state_dict (dict): scheduler state. Should be an object returned
                from a call to :meth:`state_dict`.
        """
        self.__dict__.update(state_dict)

    def step(self):
        raise NotImplementedError


class SAM(Minimizer):
    def __init__(self, model, optimizer, rho):
        super().__init__(model, optimizer)
        self.rho = rho

    def step(self, cost_fn, *inputs):
        cost = cost_fn(*inputs)
        cost.backward()
        self.ascent_step()

        self.update_records("loss_0", cost.item())

        cost = cost_fn(*inputs)
        cost.backward()
        self.descent_step()

        self.update_records("loss_p", cost.item())

    @torch.no_grad()
    def ascent_step(self):
        grad_norm = []
        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            grad_norm.append(torch.norm(p.grad, p=2))
        grad_norm = torch.norm(torch.stack(grad_norm), p=2) + 1.0e-16

        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            self.state[p]["eps"] = self.rho / grad_norm * p.grad.clone().detach()
            p.add_(self.state[p]["eps"])
        self.optimizer.zero_grad()

    @torch.no_grad()
    def descent_step(self):
        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            p.sub_(self.state[p]["eps"])
        self.optimizer.step()
        self.optimizer.zero_grad()


class ASAM(SAM):
    def __init__(self, optimizer, model, rho, eta=0.01):
        super().__init__(optimizer, model, rho)
        self.eta = eta

    @torch.no_grad()
    def ascent_step(self):
        wgrads = []
        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            t_w = self.state[p].get("eps")
            if t_w is None:
                t_w = torch.clone(p).detach()
                self.state[p]["eps"] = t_w
            if "weight" in n:
                t_w[...] = p[...]
                t_w.abs_().add_(self.eta)
                p.grad.mul_(t_w)
            wgrads.append(torch.norm(p.grad, p=2))
        wgrad_norm = torch.norm(torch.stack(wgrads), p=2) + 1.0e-16
        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            t_w = self.state[p].get("eps")
            if "weight" in n:
                p.grad.mul_(t_w)
            eps = t_w
            eps[...] = p.grad[...]
            eps.mul_(self.rho / wgrad_norm)
            p.add_(eps)
        self.optimizer.zero_grad()


class GSAM(SAM):
    def __init__(
        self,
        optimizer,
        model,
        rho,
        alpha=0.1,
        decay=True,
        rho_min=0,
        lr_max=None,
        lr_min=0,
    ):
        super().__init__(optimizer, model, rho)
        self.alpha = alpha
        self.decay = decay
        assert self.decay and (lr_max is not None)
        self.rho_min = rho_min
        self.lr_max = lr_max
        self.lr_min = lr_min

    @torch.no_grad()
    def ascent_step(self):
        grads = []
        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            self.state["ascent"][n] = p.grad.clone().detach()
            grads.append(torch.norm(p.grad, p=2))
        grad_norm = torch.norm(torch.stack(grads), p=2) + 1.0e-16
        self.ascent_norm = grad_norm.clone().detach()

        if self.decay:
            lr = self.optimizer.param_groups[0]["lr"]
            rho = self.rho_min + (self.rho - self.rho_min) * (lr - self.lr_min) / (
                self.lr_max - self.lr_min
            )
        else:
            rho = self.rho

        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            eps = self.state[p].get("eps")
            if eps is None:
                eps = torch.clone(p).detach()
                self.state[p]["eps"] = eps
            eps[...] = p.grad[...]
            eps.mul_(rho / grad_norm)
            p.add_(eps)
        self.optimizer.zero_grad()

    @torch.no_grad()
    def descent_step(self):
        self.state["descent"] = {}
        grads = []
        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            self.state["descent"][n] = p.grad.clone().detach()
            grads.append(torch.norm(p.grad, p=2))
        grad_norm = torch.norm(torch.stack(grads), p=2) + 1.0e-16
        self.descent_norm = grad_norm

        descents = self.state["descent"]
        ascents = self.state["ascent"]

        inner_prod = self.inner_product(descents, ascents)

        # get cosine
        cosine = inner_prod / (self.ascent_norm * self.descent_norm + 1.0e-16)

        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            vertical = ascents[n] - cosine * self.ascent_norm * descents[n] / (
                self.descent_norm + 1.0e-16
            )
            p.grad.sub_(self.alpha * vertical)

        self.optimizer.step()
        self.optimizer.zero_grad()

    def inner_product(self, u, v):
        value = 0
        for key in v.keys():
            value += torch.dot(u[key].flatten(), v[key].flatten())
        return value


class BridgedNSAM(SAM):
    def __init__(self, optimizer, model, rho, N=None, rhos=None, descents_weights=None):
        super().__init__(optimizer, model, rho)
        assert (N is not None) or (rhos is not None)
        if N is not None:
            self.N = N
            self.rhos = [rho / N] * N
        if rhos is not None:
            self.rhos = rhos
            self.N = len(rhos)
        if descents_weights is None:
            self.descents_weights = [1 / N] * N

    def step(self, cost_fn, *inputs):
        self.n = 0
        self.rho = self.rhos[self.n]
        cost = cost_fn(*inputs)
        cost.backward()
        self.ascent_step()

        self.update_records("loss_0", cost.item())

        for n in range(self.N - 2):
            self.n = n
            self.rho = self.rhos[self.n]
            cost = cost_fn(*inputs)
            cost.backward()
            self.middle_step()

            self.update_records("loss_%d" % n, cost.item())

        self.n = n + 1
        self.rho = self.rhos[self.n]
        cost = cost_fn(*inputs)
        cost.backward()
        self.descent_step()

        self.update_records("loss_p", cost.item())

    # Calculate First Grad and First Grad Norm. Second Ascent by First Grad.
    @torch.no_grad()
    def middle_step(self):
        grads = []
        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            self.state[p]["grad"] = (
                self.descents_weights[self.n] * p.grad.clone().detach()
            )  # Add grad
            grads.append(
                torch.norm(p.grad.clone().detach(), p=2)
            )  # Current grad to calculate norm
        grad_norm = torch.norm(torch.stack(grads), p=2) + 1.0e-16

        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            eps_t = self.rho * p.grad.clone().detach() / grad_norm
            self.state[p]["eps"] += eps_t  # Accumulate eps
            p.add_(eps_t)
        self.optimizer.zero_grad()

    @torch.no_grad()
    def descent_step(self):
        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            p.sub_(self.state[p]["eps"])  # Back ALL eps
            p.grad.mul_(self.descents_weights[self.n])
            p.grad.add_(self.state[p]["grad"])  # Add ALL grads

        self.optimizer.step()
        self.optimizer.zero_grad()


class AWP(SAM):
    @torch.no_grad()
    def ascent_step(self):
        weights = []
        grads = []
        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            weights.append(torch.norm(p, p=2))
            grads.append(torch.norm(p.grad, p=2))
        weight_norm = torch.norm(torch.stack(weights), p=2) + 1.0e-16
        grad_norm = torch.norm(torch.stack(grads), p=2) + 1.0e-16

        self.update_records("weight_norm", weight_norm.item())
        self.update_records("grad_norm", grad_norm.item())

        for n, p in self.model.named_parameters():
            if p.grad is None:
                continue
            eps = self.state[p].get("eps")
            if eps is None:
                eps = torch.clone(p).detach()
                self.state[p]["eps"] = eps
            eps[...] = p.grad[...]
            eps.mul_(self.rho * weight_norm / grad_norm)
            p.add_(eps)
        self.optimizer.zero_grad()



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).
import argparse
import copy

import torch
from airtk.attack import create_attack
from airtk.helper_functions import (
    cw_whitebox_eval,
    fgsm_whitebox_eval,
    load_data,
    mim_whitebox_eval,
)
from airtk.model import (
    ResNet18,
    ResNet34,
    ResNet50,
    ResNet101,
    ResNet152,
    WideResNet,
)
from autoattack import AutoAttack
from torch import nn
from torch.autograd import Variable
from torch.optim.swa_utils import AveragedModel

kwargs = (
    {"num_workers": 1, "pin_memory": True} if torch.cuda.is_available() else {}
)

parser = argparse.ArgumentParser(
    description="Framework for Adversarial Testing"
)

parser.add_argument(
    "--batch-size",
    type=int,
    default=128,
    metavar="N",
    help="input batch size for training (default: 128)",
)

parser.add_argument(
    "--test-batch-size",
    type=int,
    default=128,
    metavar="N",
    help="input batch size for testing (default: 128)",
)

parser.add_argument("--epsilon", default=0.031, help="perturbation")

args = parser.parse_args()


device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def run_eval(model, dataset, device, test_loader, filename, model_type):
    """

    Main function to run evaluation on a given model.


    :param model: trained image classifier to be tested

    :param dataset: dataset the model is trained on

    :param device: current device

    :param test_loader: data loader containing test set for evaluation

    :param filename: name of file to print results to

    :param model_type: base model of the trained classifier
    """

    # Open an output file and put the model in eval mode
    f = open(filename, "a")

    model.eval()

    # Set the total errors to 0 for all atts but AA
    clean_total = 0

    cw_robust_total = 0

    mim_robust_total = 0

    fgsm_robust_total = 0

    linfpgd7_robust_total = 0

    linfpgd20_robust_total = 0

    linfpgd40_robust_total = 0

    l2pgd7_robust_total = 0

    l2pgd20_robust_total = 0

    l2pgd40_robust_total = 0

    aa_robust_total = 0

    # Set the criterion to cross-entropy loss

    criterion = nn.CrossEntropyLoss()

    # Run tests for each element in the test set

    for batch_idx, (data, target) in enumerate(test_loader):
        # Set up the data/X and target/y correctly for evaluation

        data, target = data.to(device), target.to(device)

        data, target = Variable(data, requires_grad=True), Variable(target)

        # Calculate the natural and robust error for each attack

        cw_err_natural, cw_err_robust = cw_whitebox_eval(
            model, dataset, data, target, device
        )

        mim_err_natural, mim_err_robust = mim_whitebox_eval(
            model, data, target, device
        )

        fgsm_err_natural, fgsm_err_robust = fgsm_whitebox_eval(
            model, data, target
        )

        # Calculate the PGD examples and corresponding robust errors

        # l2pgd7

        attack_l2pgd7 = create_attack(
            model, criterion, "l2-pgd", 0.03, 7, 0.01
        )

        x_l2pgd7, _ = attack_l2pgd7.perturb(data, target)

        err_l2pgd7 = (
            (model(x_l2pgd7).data.max(1)[1] != target.data).float().sum()
        )

        # l2pgd20

        attack_l2pgd20 = create_attack(
            model, criterion, "l2-pgd", 0.03, 20, 0.01
        )

        x_l2pgd20, _ = attack_l2pgd20.perturb(data, target)

        err_l2pgd20 = (
            (model(x_l2pgd20).data.max(1)[1] != target.data).float().sum()
        )

        # l2pgd40

        attack_l2pgd40 = create_attack(
            model, criterion, "l2-pgd", 0.03, 40, 0.01
        )

        x_l2pgd40, _ = attack_l2pgd40.perturb(data, target)

        err_l2pgd40 = (
            (model(x_l2pgd40).data.max(1)[1] != target.data).float().sum()
        )

        # linfpgd7

        attack_linfpgd7 = create_attack(
            model, criterion, "linf-pgd", 0.03, 7, 0.01
        )

        x_linfpgd7, _ = attack_linfpgd7.perturb(data, target)

        err_linfpgd7 = (
            (model(x_linfpgd7).data.max(1)[1] != target.data).float().sum()
        )

        # linfpgd20

        attack_linfpgd20 = create_attack(
            model, criterion, "linf-pgd", 0.03, 20, 0.01
        )

        x_linfpgd20, _ = attack_linfpgd20.perturb(data, target)

        err_linfpgd20 = (
            (model(x_linfpgd20).data.max(1)[1] != target.data).float().sum()
        )

        # linfpgd40

        attack_linfpgd40 = create_attack(
            model, criterion, "linf-pgd", 0.03, 40, 0.01
        )

        x_linfpgd40, _ = attack_linfpgd40.perturb(data, target)

        err_linfpgd40 = (
            (model(x_linfpgd40).data.max(1)[1] != target.data).float().sum()
        )

        # Calculate the AA examples and corresponding robust errors

        # aa

        adversary_aa = AutoAttack(
            model, norm="Linf", eps=0.031, version="standard", verbose=False
        )

        x_aa = adversary_aa.run_standard_evaluation(data, target)

        err_aa = (model(x_aa).data.max(1)[1] != target.data).float().sum()

        # Add the losses to the total loss for each attack

        clean_total += cw_err_natural

        cw_robust_total += cw_err_robust

        mim_robust_total += mim_err_robust

        fgsm_robust_total += fgsm_err_robust

        # Add the pgd losses to the total losses

        linfpgd7_robust_total += err_linfpgd7

        linfpgd20_robust_total += err_linfpgd20

        linfpgd40_robust_total += err_linfpgd40

        l2pgd7_robust_total += err_l2pgd7

        l2pgd20_robust_total += err_l2pgd20

        l2pgd40_robust_total += err_l2pgd40

        # Add the AA losses to the total losses

        aa_robust_total += err_aa

    # Convert the clean loss tensor to accuracy %
    clean_total = int(clean_total)

    clean_acc = (10000 - clean_total) / 100

    # Convert the CW loss tensor to accuracy %

    cw_robust_total = int(cw_robust_total)

    cw_acc = (10000 - cw_robust_total) / 100

    # Convert the MIM loss tensor to accuracy %

    mim_robust_total = int(mim_robust_total)

    mim_acc = (10000 - mim_robust_total) / 100

    # Convert the FGSM loss tensor to accuracy %

    fgsm_robust_total = int(fgsm_robust_total)

    fgsm_acc = (10000 - fgsm_robust_total) / 100

    # Convert the linfpgd7 loss tensor to accuracy %

    linfpgd7_robust_total = int(linfpgd7_robust_total)

    linfpgd7_acc = (10000 - linfpgd7_robust_total) / 100

    # Convert the linfpgd20 loss tensor to accuracy %

    linfpgd20_robust_total = int(linfpgd20_robust_total)

    linfpgd20_acc = (10000 - linfpgd20_robust_total) / 100

    # Convert the linfpgd40 loss tensor to accuracy %

    linfpgd40_robust_total = int(linfpgd40_robust_total)

    linfpgd40_acc = (10000 - linfpgd40_robust_total) / 100

    # Convert the l2pgd7 loss tensor to accuracy %

    l2pgd7_robust_total = int(l2pgd7_robust_total)

    l2pgd7_acc = (10000 - l2pgd7_robust_total) / 100

    # Convert the l2pgd20 loss tensor to accuracy %

    l2pgd20_robust_total = int(l2pgd20_robust_total)

    l2pgd20_acc = (10000 - l2pgd20_robust_total) / 100

    # Convert the l2pgd40 loss tensor to accuracy %

    l2pgd40_robust_total = int(l2pgd40_robust_total)

    l2pgd40_acc = (10000 - l2pgd40_robust_total) / 100

    # Convert the aa loss tensor to accuracy %

    aa_robust_total = int(aa_robust_total)

    aa_acc = (10000 - aa_robust_total) / 100

    # Write the total losses to the file in % format

    f.write("Clean Accuracy: " + str(clean_acc) + "%\n")

    f.write("C&W Accuracy: " + str(cw_acc) + "%\n")

    f.write("MIM Accuracy: " + str(mim_acc) + "%\n")

    f.write("FGSM Accuracy: " + str(fgsm_acc) + "%\n")

    # Add the pgd losses to the total losses

    f.write("linf-pgd-7 Accuracy: " + str(linfpgd7_acc) + "%\n")

    f.write("linf-pgd-20 Accuracy: " + str(linfpgd20_acc) + "%\n")

    f.write("linf-pgd-40 Accuracy: " + str(linfpgd40_acc) + "%\n")

    f.write("l2-pgd-7 Accuracy: " + str(l2pgd7_acc) + "%\n")

    f.write("l2-pgd-20 Accuracy: " + str(l2pgd20_acc) + "%\n")

    f.write("l2-pgd-40 Accuracy: " + str(l2pgd40_acc) + "%\n")

    # Add the AA losses to the total losses

    f.write("AutoAttack Accuracy: " + str(aa_acc) + "%\n")

    f.close()  # close the file


def main_testing(dataset, base_model, filename, model_string):
    """

    Main function that sets up the testing process.


    :param dataset: dataset the tested classifier was trained on

    :param base_model: model the tested classifier was built on

    :param filename: name of file to write results to

    :param model_string: location of model to be tested
    """

    # Set up the dataset and base model

    train_loader, test_loader = load_data(dataset, args, kwargs)

    if base_model == "res18":
        model = ResNet18(num_classes=100 if dataset == "cifar100" else 10).to(
            device
        )

    elif base_model == "res34":
        model = ResNet34(num_classes=100 if dataset == "cifar100" else 10).to(
            device
        )

    elif base_model == "res50":
        model = ResNet50(num_classes=100 if dataset == "cifar100" else 10).to(
            device
        )

    elif base_model == "res101":
        model = ResNet101(num_classes=100 if dataset == "cifar100" else 10).to(
            device
        )

    elif base_model == "res152":
        model = ResNet152(num_classes=100 if dataset == "cifar100" else 10).to(
            device
        )

    elif base_model == "wideres28":
        model = WideResNet(
            depth=28, num_classes=100 if dataset == "cifar100" else 10
        ).to(device)

    elif base_model == "wideres32":
        model = WideResNet(
            depth=32, num_classes=100 if dataset == "cifar100" else 10
        ).to(device)

    elif base_model == "wideres34":
        model = WideResNet(
            depth=34, num_classes=100 if dataset == "cifar100" else 10
        ).to(device)

    elif base_model == "wideres34swa":
        model_base = WideResNet(
            depth=34, num_classes=100 if dataset == "cifar100" else 10
        ).to(device)

        model = AveragedModel(model_base).to(device)

    else:
        raise NotImplementedError

    # Load in the desired model and make a deepcopy

    model.load_state_dict(torch.load(model_string))

    model_copy = copy.deepcopy(model)

    # Run the desired evaluation on the model copy and test set

    run_eval(model_copy, dataset, device, test_loader, filename, base_model)


if __name__ == "__main__":
    main_testing(
        "cifar10",
        "wideres34swa",
        "tests/adt-va-cifar10-every9-epoch82-full-results.txt",
        "data/model/adt-va-every9/swa-model-adt-va-cifar10-wideres34-epoch82.pt",
    )



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

from airtk.attack.apgd import (
    L2APGDAttack,
    LinfAPGDAttack,
)
from airtk.attack.fgsm import FGMAttack, FGSMAttack
from airtk.attack.pgd import L2PGDAttack, LinfPGDAttack


def create_attack(
    model,
    criterion,
    attack_type,
    attack_eps,
    attack_iter,
    attack_step,
    rand_init_type="uniform",
    clip_min=0.0,
    clip_max=1.0,
):
    """

     Initialize adversary.

     Arguments:

         model (nn.Module): forward pass function.

         criterion (nn.Module): loss function.

         attack_type (str): name of the attack.

         attack_eps (float): attack radius.

         attack_iter (int): number of attack iterations.

         attack_step (float): step size for the attack.

         rand_init_type (str): random initialization type for PGD (default: uniform).

         clip_min (float): mininum value per input dimension.

         clip_max (float): maximum value per input dimension.

    Returns:

        Attack

    """

    if attack_type == "fgsm":
        attack = FGSMAttack(
            model,
            criterion,
            eps=attack_eps,
            clip_min=clip_min,
            clip_max=clip_max,
        )

    elif attack_type == "fgm":
        attack = FGMAttack(
            model,
            criterion,
            eps=attack_eps,
            clip_min=clip_min,
            clip_max=clip_max,
        )

    elif attack_type == "linf-pgd":
        attack = LinfPGDAttack(
            model,
            criterion,
            eps=attack_eps,
            nb_iter=attack_iter,
            eps_iter=attack_step,
            rand_init_type=rand_init_type,
            clip_min=clip_min,
            clip_max=clip_max,
        )

    elif attack_type == "l2-pgd":
        attack = L2PGDAttack(
            model,
            criterion,
            eps=attack_eps,
            nb_iter=attack_iter,
            eps_iter=attack_step,
            rand_init_type=rand_init_type,
            clip_min=clip_min,
            clip_max=clip_max,
        )

    elif attack_type == "pgd-targeted":
        attack = LinfPGDAttack(
            model,
            criterion,
            eps=attack_eps,
            nb_iter=attack_iter,
            eps_iter=attack_step,
            rand_init_type=rand_init_type,
            clip_min=clip_min,
            clip_max=clip_max,
            targeted=True,
        )

    elif attack_type == "linf-apgd":
        attack = LinfAPGDAttack(
            model, criterion, n_restarts=2, eps=attack_eps, nb_iter=attack_iter
        )

    elif attack_type == "l2-apgd":
        attack = L2APGDAttack(
            model, criterion, n_restarts=2, eps=attack_eps, nb_iter=attack_iter
        )

    else:
        raise NotImplementedError(f"{attack_type} is not yet implemented!")

    return attack



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
# A simple torch style logger
# (C) Wei YANG 2017

import numpy as np

__all__ = ["Logger", "LoggerMonitor", "savefig"]


def savefig(fname, dpi=None):
    dpi = 150 if dpi is None else dpi
    # plt.savefig(fname, dpi=dpi)


def plot_overlap(logger, names=None):
    names = logger.names if names is None else names
    numbers = logger.numbers
    for _, name in enumerate(names):
        x = np.arange(len(numbers[name]))
        # plt.plot(x, np.asarray(numbers[name]))
    return [logger.title + "(" + name + ")" for name in names]


class Logger:
    """Save training process to log file with simple plot function."""

    def __init__(self, fpath, title=None, resume=False):
        self.file = None
        self.resume = resume
        self.title = "" if title == None else title
        if fpath is not None:
            if resume:
                self.file = open(fpath)
                name = self.file.readline()
                self.names = name.rstrip().split("\t")
                self.numbers = {}
                for _, name in enumerate(self.names):
                    self.numbers[name] = []

                for numbers in self.file:
                    numbers = numbers.rstrip().split("\t")
                    for i in range(0, len(numbers)):
                        self.numbers[self.names[i]].append(numbers[i])
                self.file.close()
                self.file = open(fpath, "a")
            else:
                self.file = open(fpath, "w")

    def set_names(self, names):
        if self.resume:
            pass
        # initialize numbers as empty list
        self.numbers = {}
        self.names = names
        for _, name in enumerate(self.names):
            self.file.write(name)
            self.file.write("\t")
            self.numbers[name] = []
        self.file.write("\n")
        self.file.flush()

    def append(self, numbers):
        assert len(self.names) == len(numbers), "Numbers do not match names"
        for index, num in enumerate(numbers):
            self.file.write(f"{num:.6f}")
            self.file.write("\t")
            self.numbers[self.names[index]].append(num)
        self.file.write("\n")
        self.file.flush()

    def plot(self, names=None):
        names = self.names if names == None else names
        numbers = self.numbers
        for _, name in enumerate(names):
            x = np.arange(len(numbers[name]))
            # plt.plot(x, np.asarray(numbers[name]))
        # plt.legend([self.title + '(' + name + ')' for name in names])
        # plt.grid(True)

    def close(self):
        if self.file is not None:
            self.file.close()


class LoggerMonitor:
    """Load and visualize multiple logs."""

    def __init__(self, paths):
        """paths is a distionary with {name:filepath} pair"""
        self.loggers = []
        for title, path in paths.items():
            logger = Logger(path, title=title, resume=True)
            self.loggers.append(logger)

    def plot(self, names=None):
        # plt.figure()
        # plt.subplot(121)
        legend_text = []
        for logger in self.loggers:
            legend_text += plot_overlap(logger, names)
        # plt.legend(legend_text, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        # plt.grid(True)


if __name__ == "__main__":
    # # Example
    # logger = Logger('test.txt')
    # logger.set_names(['Train loss', 'Valid loss','Test loss'])

    # length = 100
    # t = np.arange(length)
    # train_loss = np.exp(-t / 10.0) + np.random.rand(length) * 0.1
    # valid_loss = np.exp(-t / 10.0) + np.random.rand(length) * 0.1
    # test_loss = np.exp(-t / 10.0) + np.random.rand(length) * 0.1

    # for i in range(0, length):
    #     logger.append([train_loss[i], valid_loss[i], test_loss[i]])
    # logger.plot()

    # Example: logger monitor
    paths = {
        "resadvnet20": "/home/wyang/code/pytorch-classification/checkpoint/cifar10/resadvnet20/log.txt",
        "resadvnet32": "/home/wyang/code/pytorch-classification/checkpoint/cifar10/resadvnet32/log.txt",
        "resadvnet44": "/home/wyang/code/pytorch-classification/checkpoint/cifar10/resadvnet44/log.txt",
    }

    field = ["Valid Acc."]

    monitor = LoggerMonitor(paths)
    monitor.plot(names=field)
    savefig("test.eps")



"""
Reference:
[1] Towards Deep Learning Models Resistant to Adversarial Attacks
Aleksander Madry, Aleksandar Makelov, Ludwig Schmidt, Dimitris Tsipras, Adrian Vladu
arXiv:1706.06083v3
"""

import os
import sys

import numpy as np
import torch

father_dir = os.path.join(
    "/", *os.path.realpath(__file__).split(os.path.sep)[:-2]
)
if father_dir not in sys.path:
    sys.path.append(father_dir)
from airtk.defense.yopo.attack_base import (
    AttackBase,
    clip_eta,
)


class IPGD(AttackBase):
    # ImageNet pre-trained mean and std
    # _mean = torch.tensor(np.array([0.485, 0.456, 0.406]).astype(np.float32)[np.newaxis, :, np.newaxis, np.newaxis])
    # _std = torch.tensor(np.array([0.229, 0.224, 0.225]).astype(np.float32)[np.newaxis, :, np.newaxis, np.newaxis])

    # _mean = torch.tensor(np.array([0]).astype(np.float32)[np.newaxis, :, np.newaxis, np.newaxis])
    # _std = torch.tensor(np.array([1.0]).astype(np.float32)[np.newaxis, :, np.newaxis, np.newaxis])
    def __init__(
        self,
        eps=6 / 255.0,
        sigma=3 / 255.0,
        nb_iter=20,
        norm=np.inf,
        DEVICE=torch.device("cpu"),
        mean=torch.tensor(
            np.array([0]).astype(np.float32)[
                np.newaxis, :, np.newaxis, np.newaxis
            ]
        ),
        std=torch.tensor(
            np.array([1.0]).astype(np.float32)[
                np.newaxis, :, np.newaxis, np.newaxis
            ]
        ),
        random_start=True,
    ):
        """
        :param eps: maximum distortion of adversarial examples
        :param sigma: single step size
        :param nb_iter: number of attack iterations
        :param norm: which norm to bound the perturbations
        """
        self.eps = eps
        self.sigma = sigma
        self.nb_iter = nb_iter
        self.norm = norm
        self.criterion = torch.nn.CrossEntropyLoss().to(DEVICE)
        self.DEVICE = DEVICE
        self._mean = mean.to(DEVICE)
        self._std = std.to(DEVICE)
        self.random_start = random_start

    def single_attack(self, net, inp, label, eta, target=None):
        """
        Given the original image and the perturbation computed so far, computes
        a new perturbation.
        :param net:
        :param inp: original image
        :param label:
        :param eta: perturbation computed so far
        :return: a new perturbation
        """

        adv_inp = inp + eta

        # net.zero_grad()

        pred = net(adv_inp)
        if target is not None:
            targets = torch.sum(pred[:, target])
            grad_sign = torch.autograd.grad(
                targets, adv_inp, only_inputs=True, retain_graph=False
            )[0].sign()  # was adv_in

        else:
            loss = self.criterion(pred, label)
            grad_sign = torch.autograd.grad(
                loss, adv_inp, only_inputs=True, retain_graph=False
            )[0].sign()

        adv_inp = adv_inp + grad_sign * (self.sigma / self._std)
        tmp_adv_inp = adv_inp * self._std + self._mean

        tmp_inp = inp * self._std + self._mean
        tmp_adv_inp = torch.clamp(tmp_adv_inp, 0, 1)  ## clip into 0-1
        # tmp_adv_inp = (tmp_adv_inp - self._mean) / self._std
        tmp_eta = tmp_adv_inp - tmp_inp
        tmp_eta = clip_eta(
            tmp_eta, norm=self.norm, eps=self.eps, DEVICE=self.DEVICE
        )

        eta = tmp_eta / self._std

        return eta

    def attack(self, net, inp, label, target=None):
        if self.random_start:
            eta = torch.FloatTensor(*inp.shape).uniform_(-self.eps, self.eps)
        else:
            eta = torch.zeros_like(inp)
        eta = eta.to(self.DEVICE)
        eta = (eta - self._mean) / self._std
        net.eval()

        inp.requires_grad = True
        eta.requires_grad = True
        for i in range(self.nb_iter):
            eta = self.single_attack(net, inp, label, eta, target)
            # print(i)

        # print(eta.max())
        adv_inp = inp + eta
        tmp_adv_inp = adv_inp * self._std + self._mean
        tmp_adv_inp = torch.clamp(tmp_adv_inp, 0, 1)
        adv_inp = (tmp_adv_inp - self._mean) / self._std

        return adv_inp

    def to(self, device):
        self.DEVICE = device
        self._mean = self._mean.to(device)
        self._std = self._std.to(device)
        self.criterion = self.criterion.to(device)


def test_IPGD():
    pass


if __name__ == "__main__":
    test_IPGD()



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

import pickle

import torch
from torch import nn

# from torch.autograd.gradcheck import zero_gradients
from torch.autograd import Variable

from airtk.defense.feature_scatter import fs_ot
from airtk.defense.feature_scatter.fs_utils import (
    label_smoothing,
    one_hot_tensor,
    softCrossEntropy,
)

device = "cuda" if torch.cuda.is_available() else "cpu"


def zero_gradients(x):
    """Manually zeroes the gradients when pytorch is unable"""

    if x.grad is not None:
        x.grad.zero_()


class Attack_FeaScatter(nn.Module):
    def __init__(self, basic_net, config, attack_net=None):
        super().__init__()

        self.basic_net = basic_net

        self.attack_net = attack_net

        self.rand = config["random_start"]

        self.step_size = config["step_size"]

        self.epsilon = config["epsilon"]

        self.num_steps = config["num_steps"]
        self.train_flag = (
            True if "train" not in config.keys() else config["train"]
        )

        self.box_type = (
            "white" if "box_type" not in config.keys() else config["box_type"]
        )
        self.ls_factor = (
            0.1 if "ls_factor" not in config.keys() else config["ls_factor"]
        )

        print(config)

    def forward(
        self, inputs, targets, attack=True, targeted_label=-1, batch_idx=0
    ):
        if not attack:
            outputs, _ = self.basic_net(inputs, return_feature=True)

            return outputs, None

        if self.box_type == "white":
            aux_net = pickle.loads(pickle.dumps(self.basic_net))

        elif self.box_type == "black":
            assert (
                self.attack_net is not None
            ), "should provide an additional net in black-box case"

            aux_net = pickle.loads(pickle.dumps(self.basic_net))

        aux_net.eval()

        batch_size = inputs.size(0)

        m = batch_size

        n = batch_size

        logits = aux_net(inputs, return_feature=True)[0]

        num_classes = logits.size(1)

        outputs = aux_net(inputs, return_feature=True)[0]

        x = inputs.detach()

        x = x + torch.zeros_like(x).uniform_(-self.epsilon, self.epsilon)

        if self.train_flag:
            self.basic_net.train()

        else:
            self.basic_net.eval()

        logits_pred_nat, fea_nat = aux_net(inputs, return_feature=True)

        num_classes = logits_pred_nat.size(1)

        y_gt = one_hot_tensor(targets, num_classes, device)

        loss_ce = softCrossEntropy()

        iter_num = self.num_steps

        for _ in range(iter_num):
            x.requires_grad_()

            zero_gradients(x)

            if x.grad is not None:
                x.grad.data.fill_(0)

            logits_pred, fea = aux_net(x, return_feature=True)

            ot_loss = fs_ot.sinkhorn_loss_joint_IPOT(
                1, 0.00, logits_pred_nat, logits_pred, None, None, 0.01, m, n
            )

            aux_net.zero_grad()

            adv_loss = ot_loss

            adv_loss.backward(retain_graph=True)

            x_adv = x.data + self.step_size * torch.sign(x.grad.data)

            x_adv = torch.min(
                torch.max(x_adv, inputs - self.epsilon), inputs + self.epsilon
            )

            x_adv = torch.clamp(x_adv, -1.0, 1.0)

            x = Variable(x_adv)

            logits_pred, fea = self.basic_net(x, return_feature=True)

            self.basic_net.zero_grad()

            y_sm = label_smoothing(y_gt, y_gt.size(1), self.ls_factor)

            adv_loss = loss_ce(logits_pred, y_sm.detach())

        return logits_pred, adv_loss



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).

from airtk import (
    attack,
    data,
    defense,
    loss,
    model,
    util,
)

__all__ = ["attack", "defense", "model", "data", "loss", "util"]



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).


import copy
import os
import time
from collections import namedtuple

import torch
from torch import optim
from torch.nn import Module

from airtk.helper_functions import (
    adjust_learning_rate,
    eval_clean,
    load_data,
    robust_eval,
)
from airtk.loss import adt_loss
from airtk.model import (
    ResNet18,
    ResNet34,
    ResNet50,
    ResNet101,
    ResNet152,
    WideResNet,
)


class AdtTraining:
    def __init__(
        self,
        ds_name: str,
        model_name: str,
        batch_size: int = 128,
        epochs: int = 76,
        weight_decay: float = 2e-4,
        lr: float = 0.1,
        momentum: float = 0.9,
        seed: int = 0,
        model_dir: str = "data/model",
        lr_schedule: str = "decay",
    ) -> None:
        ArgsPrototype = namedtuple(
            "Prototype",
            [
                "batch_size",
                "test_batch_size",
                "epochs",
                "weight_decay",
                "lr",
                "momentum",
                "seed",
                "model_dir",
                "save_freq",
                "lr_schedule",
                "log_interval",
            ],
        )

        self._args = ArgsPrototype(
            batch_size,
            batch_size,
            epochs,
            weight_decay,
            lr,
            momentum,
            seed,
            model_dir,
            1,
            seed,
            1,
        )

        self._ds_name: str = ds_name

        self._model_name: str = model_name

    def __call__(self) -> Module:
        model_dir: str = self._args.model_dir

        if not os.path.exists(model_dir):
            os.makedirs(model_dir)

        torch.manual_seed(self._args.seed)

        main_adt(self._ds_name, self._model_name, self._args)


def train(args, model, device, train_loader, optimizer, ds_name, epoch):
    """

    Trains one epoch by calculating the loss for each batch and updating the model


    :param args: set of arguments including learning rate, log interval, etc

    :param model: model to train

    :param device: the current device

    :param train_loader: data loader containing the training dataset

    :param optimizer: optimizer to train

    :param epoch: current epoch of training
    """

    model.train()  # set the  model to training mode

    for batch_idx, sample in enumerate(train_loader):
        # Set the data and target (x and y) to the device

        if ds_name == "cifar100":
            data, target, coarse = sample

        else:
            data, target = sample

        data, target = data.to(device), target.to(device)

        optimizer.zero_grad()  # zero out the gradients

        # Calculate the ADT loss

        loss = adt_loss(model, data, target, optimizer)

        # Update the model based on the loss

        loss.backward()

        optimizer.step()

        # print progress

        if batch_idx % args.log_interval == 0:
            print(
                f"Train Epoch: {epoch} [{batch_idx * len(data)}/{len(train_loader.dataset)} ({100.0 * batch_idx / len(train_loader):.0f}%)]\tLoss: {loss.item():.6f}"
            )


def main_adt(ds_name, mod_name, args):
    """

    Main training method, which establishes the model/dataset before conducting training and

    clean testing for each epoch. Then reports final training time and conducts robustness tests
    on the final model.


    :param ds_name: the dataset to use for training

    :param mod_name: the model to use for training
    """

    # Set up file for printing the output

    filename = f"log/adt-{ds_name}-{mod_name}-output.txt"
    f = open(filename, "a")

    device = "cuda" if torch.cuda.is_available() else "cpu"

    # Initialize the desired model

    start_tot = time.time()  # start timing the full method

    if mod_name == "res18":
        model = ResNet18(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res34":
        model = ResNet34(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res50":
        model = ResNet50(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res101":
        model = ResNet101(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res152":
        model = ResNet152(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "wideres34":
        model = WideResNet(
            depth=34, num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    else:
        raise NotImplementedError

    # Set up the optimizer to the arguments

    optimizer = optim.SGD(
        model.parameters(),
        lr=args.lr,
        momentum=args.momentum,
        weight_decay=args.weight_decay,
    )

    # Set up the dataloaders

    kwargs = (
        {"num_workers": 1, "pin_memory": True}
        if torch.cuda.is_available()
        else {}
    )

    train_loader, test_loader = load_data(ds_name, args, kwargs, coarse=True)

    # Save the model and optimizer

    torch.save(
        model.state_dict(),
        os.path.join(
            args.model_dir, f"model-adt-{ds_name}-{mod_name}-start.pt"
        ),
    )

    torch.save(
        optimizer.state_dict(),
        os.path.join(
            args.model_dir, f"opt-adt-{ds_name}-{mod_name}-start.tar"
        ),
    )

    # Begin training for the designated number of epochs

    for epoch in range(1, args.epochs + 1):
        # Adjust the learning rate for the SGD

        adjust_learning_rate(optimizer, epoch, args)

        # Commence training, timing the method and recording the time

        start = time.time()  # start timing the method

        train(args, model, device, train_loader, optimizer, ds_name, epoch)

        end = time.time()  # finish timing the method

        epoch_time = end - start

        # evaluation on natural examples (training and testing)
        print(
            "================================================================"
        )

        f.write("Epoch " + str(epoch) + "\n")

        print("Time for Training: " + str(epoch_time))

        f.write("Time for Training: " + str(epoch_time) + "\n")

        eval_clean(model, device, train_loader, "train", ds_name, f)

        eval_clean(model, device, test_loader, "test", ds_name, f)

        robust_eval(model, device, test_loader, ds_name, f)
        print(
            "================================================================"
        )

        # Save the model (if designated)

        if epoch % args.save_freq == 0:
            torch.save(
                model.state_dict(),
                os.path.join(
                    args.model_dir,
                    f"model-adt-{ds_name}-{mod_name}-epoch{epoch}.pt",
                ),
            )

            torch.save(
                optimizer.state_dict(),
                os.path.join(
                    args.model_dir,
                    f"opt-adt-{ds_name}-{mod_name}-epoch{epoch}.tar",
                ),
            )

    # Report the time taken to fully train the model
    end_tot = time.time()

    total_time = end_tot - start_tot

    print("Total training time: " + str(total_time))

    f.write("Total training time: " + str(total_time) + "\n")

    # Load the final trained model, make a deepcopy, and evaluate against FGSM and PGD

    if mod_name == "res18":
        trained_model = ResNet18(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    elif mod_name == "res34":
        trained_model = ResNet34(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    elif mod_name == "res50":
        trained_model = ResNet50(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    elif mod_name == "res101":
        trained_model = ResNet101(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    elif mod_name == "res152":
        trained_model = ResNet152(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    elif mod_name == "wideres34":
        trained_model = WideResNet(
            depth=34, num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    else:
        raise NotImplementedError
    trained_model.load_state_dict(
        torch.load(
            f"data/model/model-adt-{ds_name}-{mod_name}-epoch{args.epochs}.pt"
        )
    )

    model_copy = copy.deepcopy(trained_model)

    robust_eval(model_copy, device, test_loader, ds_name, f)

    f.close()  # close the output file



import torch
import torch.nn as nn
import torch.nn.functional as F

from ..attack import Attack


class DIFGSM(Attack):
    r"""
    DI2-FGSM in the paper 'Improving Transferability of Adversarial Examples with Input Diversity'
    [https://arxiv.org/abs/1803.06978]

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)
        alpha (float): step size. (Default: 2/255)
        decay (float): momentum factor. (Default: 0.0)
        steps (int): number of iterations. (Default: 10)
        resize_rate (float): resize factor used in input diversity. (Default: 0.9)
        diversity_prob (float) : the probability of applying input diversity. (Default: 0.5)
        random_start (bool): using random initialization of delta. (Default: False)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.DIFGSM(model, eps=8/255, alpha=2/255, steps=10, decay=0.0, resize_rate=0.9, diversity_prob=0.5, random_start=False)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        eps=8 / 255,
        alpha=2 / 255,
        steps=10,
        decay=0.0,
        resize_rate=0.9,
        diversity_prob=0.5,
        random_start=False,
    ):
        super().__init__("DIFGSM", model)
        self.eps = eps
        self.steps = steps
        self.decay = decay
        self.alpha = alpha
        self.resize_rate = resize_rate
        self.diversity_prob = diversity_prob
        self.random_start = random_start
        self.supported_mode = ["default", "targeted"]

    def input_diversity(self, x):
        img_size = x.shape[-1]
        img_resize = int(img_size * self.resize_rate)

        if self.resize_rate < 1:
            img_size = img_resize
            img_resize = x.shape[-1]

        rnd = torch.randint(
            low=img_size, high=img_resize, size=(1,), dtype=torch.int32
        )
        rescaled = F.interpolate(
            x, size=[rnd, rnd], mode="bilinear", align_corners=False
        )
        h_rem = img_resize - rnd
        w_rem = img_resize - rnd
        pad_top = torch.randint(
            low=0, high=h_rem.item(), size=(1,), dtype=torch.int32
        )
        pad_bottom = h_rem - pad_top
        pad_left = torch.randint(
            low=0, high=w_rem.item(), size=(1,), dtype=torch.int32
        )
        pad_right = w_rem - pad_left

        padded = F.pad(
            rescaled,
            [
                pad_left.item(),
                pad_right.item(),
                pad_top.item(),
                pad_bottom.item(),
            ],
            value=0,
        )

        return padded if torch.rand(1) < self.diversity_prob else x

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        loss = nn.CrossEntropyLoss()
        momentum = torch.zeros_like(images).detach().to(self.device)

        adv_images = images.clone().detach()

        if self.random_start:
            # Starting at a uniformly random point
            adv_images = adv_images + torch.empty_like(adv_images).uniform_(
                -self.eps, self.eps
            )
            adv_images = torch.clamp(adv_images, min=0, max=1).detach()

        for _ in range(self.steps):
            adv_images.requires_grad = True
            outputs = self.get_logits(self.input_diversity(adv_images))

            # Calculate loss
            if self.targeted:
                cost = -loss(outputs, target_labels)
            else:
                cost = loss(outputs, labels)

            # Update adversarial images
            grad = torch.autograd.grad(
                cost, adv_images, retain_graph=False, create_graph=False
            )[0]

            grad = grad / torch.mean(
                torch.abs(grad), dim=(1, 2, 3), keepdim=True
            )
            grad = grad + momentum * self.decay
            momentum = grad

            adv_images = adv_images.detach() + self.alpha * grad.sign()
            delta = torch.clamp(
                adv_images - images, min=-self.eps, max=self.eps
            )
            adv_images = torch.clamp(images + delta, min=0, max=1).detach()

        return adv_images



import torch
import torch.nn as nn

model_info = {
    "VGG11": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "VGG13": [64, 64, "M", 128, 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "VGG16": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        "M",
        512,
        512,
        512,
        "M",
    ],
    "VGG19": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        512,
        "M",
        512,
        512,
        512,
        512,
        "M",
    ],
}


def _make_layers(model_name):
    layers = []
    in_channel = 3

    for out_channel in model_info[model_name]:
        if out_channel == "M":
            layers.append(nn.MaxPool2d(2, 2))
        else:
            layers.append(nn.Conv2d(in_channel, out_channel, kernel_size=3, padding=1))
            layers.append(nn.BatchNorm2d(out_channel))
            layers.append(nn.ReLU(inplace=True))
            in_channel = out_channel

    return layers


class VGG(nn.Module):
    def __init__(self, name, num_classes=10):
        super(VGG, self).__init__()

        self.conv_layer = nn.Sequential(*_make_layers(name))

        self.fc_layer = nn.Sequential(nn.Linear(512, num_classes))

    def forward(self, x):
        out = self.conv_layer(x)
        out = out.view(-1, 512)
        out = self.fc_layer(out)

        return out



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

import copy
import os
import time
from collections import namedtuple
from typing import Any, Dict, NamedTuple

import numpy as np
import torch
from torch import optim
from torch.nn import Module
from torch.optim import Optimizer
from torch.optim.lr_scheduler import CosineAnnealingLR
from torch.optim.swa_utils import SWALR, AveragedModel
from torch.utils.data import DataLoader

from airtk.helper_functions import (
    class_define_attacks,
    eval_clean,
    get_model,
    load_data,
    robust_eval,
)
from airtk.loss import adt_va_loss


class AdtppTraining:
    """Class which provides a simple interface for the ADT++ training method."""

    def __init__(
        self,
        dataset: str,
        model: str,
        batch_size: int = 128,
        epochs: int = 200,
        weight_decay: float = 2e-4,
        lr: float = 0.1,
        momentum: float = 0.9,
        seed: int = 1,
        log_interval: int = 1,
        model_dir: str = "data/model",
        swa_warmup: int = 50,
        swa: bool = True,
        lr_schedule: str = "decay",
    ) -> None:
        """Creates a new `AdtppTraining` instance.


        :param dataset: the dataset to be used as a str. Should be either

            "cifar10" or "cifar100".

        :param model: the model to train. Should be one of

            "res18", "res34", "res50", "res101", "res152", or "wideres34".

        :param batch_size: the size of batches for training and testing.

            Defaults to 128.

        :param epochs: the number of epochs to train for.

            Defaults to 200.

        :param weight_decay: a weight decay constant for regularization.

            Defaults to 2e-4.

        :param lr: the initial learning rate. Defaults to 0.1.

        :param momentum: Nesterov momentum constant. Defaults to 0.9.

        :param seed: the random seed to use for training. Defaults to 1.

        :param log_interval: the number of batches to log per. Defaults to 1.

        :param model_dir: the the directory to output models to.

            Defaults to "data/model".

        :param swa_warmup: the number of epochs to apply

            stochastic weight averaging to. Defaults to 50.

        :param swa: whether or not to apply stochastic weight averaging.

            Defaults to True.

        :param lr_schedule: the learning rate schedule. Should be one of

            "decay", "schedules", "cosine", "cyclic-10", or "cyclic-5".

            Defaults to "decay".
        """

        ArgsPrototype = namedtuple(
            "ArgsPrototype",
            [
                "batch_size",
                "test_batch_size",
                "epochs",
                "weight_decay",
                "lr",
                "momentum",
                "no_cuda",
                "seed",
                "log_interval",
                "model_dir",
                "swa_warmup",
                "swa",
                "lr_schedule",
                "save_freq",
            ],
        )

        self._args = ArgsPrototype(
            batch_size,
            batch_size,
            epochs,
            weight_decay,
            lr,
            momentum,
            False,
            seed,
            log_interval,
            model_dir,
            swa_warmup,
            swa,
            lr_schedule,
            1,
        )
        self._dataset = dataset
        self._model = model

    def __call__(self) -> None:
        """Trains the model defined by __init__."""

        main_va_epochs(self._dataset, self._model, self._args, 0)


def train(
    args: NamedTuple,
    model: Module,
    device: torch.device,
    train_loader: DataLoader,
    optimizer: Optimizer,
    epoch: int,
    attacks: Dict[int, str],
    ds_name: str,
    f: str,
    swa_model: Module,
    swa_scheduler: Any,
    scheduler: Any,
):
    """

    Trains one epoch (not the first) by calculating the loss for each batch

    based on the given set of atts, and updating the model


    :param args: set of arguments including learning rate, log interval, etc

    :param model: model to train

    :param device: the current device

    :param train_loader: data loader containing the training dataset

    :param optimizer: optimizer to train

    :param epoch: current epoch of training

    :param attacks: dictionary of attack types based on class

    :param ds_name: name of dataset
    """

    model.train()  # Set the model to training mode

    # For each batch of images in the train loader (size 128)

    for batch_idx, sample in enumerate(train_loader):
        # Extract the image and classification details to determine the class

        if ds_name == "cifar100":
            data, target, coarse = sample

            data, target = (
                data.to(device, non_blocking=True),
                target.to(device, non_blocking=True),
            )

            avg_cls = int(torch.mode(coarse)[0])

        else:
            data, target = sample

            data, target = (
                data.to(device, non_blocking=True),
                target.to(device, non_blocking=True),
            )

            avg_cls = int(torch.mode(target)[0])

        # Determine the attack based on the most common class

        attack = attacks[avg_cls]

        # Zero out the gradients of the optimizer

        optimizer.zero_grad()

        # Determine the loss to the batch based on the attack

        loss = adt_va_loss(
            model, data, target, attack, ds_name, optimizer, device
        )

        # Update the model parameters based on the calculated class

        loss.backward()

        optimizer.step()

        # Print the training progress

        if batch_idx % args.log_interval == 0:
            print(
                f"Train Epoch: {epoch} [{batch_idx * len(data)}"
                + f"/{len(train_loader.dataset)} "
                + f"({100.0 * batch_idx / len(train_loader):.0f}%)]\t"
                + f" Loss: {loss.item():.6f}"
            )

            f.write(
                f"Train Epoch: {epoch} [{batch_idx * len(data)}"
                + f"/{len(train_loader.dataset)} "
                + f"({100.0 * batch_idx / len(train_loader):.0f}%)]\t"
                + f"Loss: {loss.item():.6f}"
            )

    # Update the SWA model/scheduler

    if args.swa:
        if epoch > args.swa_warmup:
            swa_model.update_parameters(model)

            swa_scheduler.step()

        else:
            scheduler.step()


def main_va_epochs(
    ds_name: str, mod_name: str, args: NamedTuple, clean_epochs: int = 0
):
    """

    Main training method, which establishes the model/dataset before

    conducting training and clean testing for each epoch. Then reports final

    training time and conducts robustness tests on the final model.


    :param ds_name: the dataset to use for training

    :param mod_name: the model to use for training
    """

    # Establish the settings
    model_dir = args.model_dir

    if not os.path.exists(model_dir):
        os.makedirs(model_dir)

    use_cuda = not args.no_cuda and torch.cuda.is_available()

    torch.manual_seed(args.seed)

    device = torch.device("cuda" if use_cuda else "cpu")

    kwargs = {"num_workers": 1, "pin_memory": True} if use_cuda else {}

    # Set up file for printing the outut

    filename = f"log/adt-va-{ds_name}-{mod_name}-output.txt"

    with open(filename, "a") as f:
        # Initialize the model based on the specified parameter

        start_tot = time.time()  # start recording training time

        base_model = get_model(mod_name, ds_name, device)

        # Clean training/warmup round, if specified

        if clean_epochs == 0:
            model = copy.deepcopy(base_model).to(device)

        else:
            main_warmup(ds_name, mod_name, clean_epochs)

            base_model.load_state_dict(
                torch.load(
                    os.path.join(
                        model_dir,
                        f"model-warmup-{ds_name}-{mod_name}-epoch{clean_epochs}.pt",
                    )
                )
            )

            model = copy.deepcopy(base_model).to(device)

        # Set up the optimizer to the arguments

        optimizer = optim.SGD(
            model.parameters(),
            lr=args.lr,
            momentum=args.momentum,
            weight_decay=args.weight_decay,
        )

        if args.swa:
            swa_model = AveragedModel(model).to(device)

            scheduler = CosineAnnealingLR(optimizer, T_max=100)

            swa_scheduler = SWALR(optimizer, swa_lr=0.05)

        else:
            swa_model = model

            scheduler = optimizer

            swa_scheduler = optimizer

        # Set up the dataloaders

        train_loader, test_loader = load_data(
            ds_name, args, kwargs, coarse=True
        )

        # Save the model and optimizer

        torch.save(
            model.state_dict(),
            os.path.join(
                model_dir, f"model-adt-va-{ds_name}-{mod_name}-start.pt"
            ),
        )

        torch.save(
            optimizer.state_dict(),
            os.path.join(
                model_dir, f"opt-adt-va-{ds_name}-{mod_name}-start.tar"
            ),
        )

        # For each epoch

        for epoch in range(1, args.epochs + 1):
            # Adjust the SGD learning rate based on the epoch

            adjust_learning_rate(optimizer, epoch, args)

            # Define the dictionary of attack types

            attacks = class_define_attacks(ds_name)

            # Calculate the time elapsed

            # Commence training, timing the method and recording the time

            start = time.time()  # start recording training time

            train(
                args,
                model,
                device,
                train_loader,
                optimizer,
                epoch,
                attacks,
                ds_name,
                f,
                swa_model,
                swa_scheduler,
                scheduler,
            )

            end = time.time()  # stop recording training time

            epoch_time = end - start

            # Update bn statistics for the swa_model at the end

            if args.swa:
                torch.optim.swa_utils.update_bn(
                    train_loader, swa_model, device=device
                )

            # evaluation on natural examples (training and testing)

            print(
                "================================================================"
            )

            f.write("Epoch " + str(epoch) + "\n")

            f.write("Current Attack Dictionary: ")

            f.write(str(attacks) + "\n")

            print("Time for Training: " + str(epoch_time))

            f.write("Time for Training: " + str(epoch_time) + "\n")

            if args.swa:
                if epoch > args.swa_warmup:
                    eval_clean(
                        swa_model, device, train_loader, "train", ds_name, f
                    )

                    eval_clean(
                        swa_model, device, test_loader, "test", ds_name, f
                    )

                    robust_eval(swa_model, device, test_loader, ds_name, f)

            else:
                eval_clean(model, device, train_loader, "train", ds_name, f)

                eval_clean(model, device, test_loader, "test", ds_name, f)

                robust_eval(model, device, test_loader, ds_name, f)

            print(
                "================================================================"
            )

            # Save the model (if designated)

            if epoch % args.save_freq == 0:
                torch.save(
                    model.state_dict(),
                    os.path.join(
                        model_dir,
                        f"model-adt-va-{ds_name}-{mod_name}-epoch{epoch}.pt",
                    ),
                )

                torch.save(
                    swa_model.state_dict(),
                    os.path.join(
                        model_dir,
                        f"swa-model-adt-va-{ds_name}-{mod_name}-epoch{epoch}.pt",
                    ),
                )

                torch.save(
                    optimizer.state_dict(),
                    os.path.join(
                        model_dir,
                        f"opt-adt-va-{ds_name}-{mod_name}-epoch{epoch}.tar",
                    ),
                )

        # Report the time taken to fully train the model

        end_tot = time.time()

        total_time = end_tot - start_tot

        print("Total training time: " + str(total_time))

        f.write("Total training time: " + str(total_time) + "\n")

        # Load the final trained model, make a deepcopy, and evaluate against

        # FGSM and PGD

        trained_model = get_model(mod_name, ds_name, device)

        if args.swa:
            trained_model = AveragedModel(trained_model).to(device)

            trained_model.load_state_dict(
                torch.load(
                    os.path.join(
                        model_dir,
                        f"swa-model-adt-va-{ds_name}-{mod_name}-epoch{args.epochs}.pt",
                    )
                )
            )

        else:
            trained_model.load_state_dict(
                torch.load(
                    os.path.join(
                        model_dir,
                        f"model-adt-va-{ds_name}-{mod_name}-epoch{args.epochs}.pt",
                    )
                )
            )

        model_copy = copy.deepcopy(trained_model)

        robust_eval(model_copy, device, test_loader, ds_name, f)


def main_warmup(
    ds_name: str, mod_name: str, epochs: int, args: NamedTuple
) -> None:
    """

    Main training method, which establishes the model/dataset before

    conducting training and clean testing for each epoch. Then reports

    final training time and conducts robustness tests on the final model.


    :param ds_name: dataset to use for training

    :param mod_name: model to use for training

    :param epochs: number of epochs to train for
    """

    device = "cuda" if torch.cuda.is_available else "cpu"
    model_dir = args.model_dir

    # Create file to print training progress

    filename = f"log/clean-{ds_name}-{mod_name}-output.txt"

    kwargs = (
        {"num_workers": 1, "pin_memory": True}
        if torch.cuda.is_available()
        else {}
    )

    with open(filename, "a") as f:
        # Initialize the model based on the specified architecture

        start_tot = time.time()  # start recording training time

        model = get_model(mod_name, ds_name, device)

        # Set up the optimizer with the arguments

        optimizer = optim.SGD(
            model.parameters(),
            lr=args.lr,
            momentum=args.momentum,
            weight_decay=args.weight_decay,
        )

        # Set up the dataloaders

        train_loader, test_loader = load_data(
            ds_name, args, kwargs, coarse=True
        )

        # Begin model training for the specified number of epochs

        for epoch in range(1, epochs + 1):
            # Adjust the SGD learning rate based on the epoch

            adjust_learning_rate(optimizer, epoch, args)

            start = time.time()  # start recording training time

            train(
                args,
                model,
                device,
                train_loader,
                optimizer,
                epoch,
                ds_name,
            )

            end = time.time()  # stop recording training time

            epoch_time = end - start

            # Evaluation on natural and adversarial examples

            print(
                "================================================================"
            )

            f.write("Epoch " + str(epoch) + "\n")

            print("Time for Training: " + str(epoch_time))

            f.write("Time for Training: " + str(epoch_time) + "\n")

            eval_clean(model, device, train_loader, "train", ds_name, f)

            eval_clean(model, device, test_loader, "test", ds_name, f)

            robust_eval(model, device, test_loader, ds_name, f)

            print(
                "================================================================"
            )

            # Save the model (if designated)

            if epoch % args.save_freq == 0:
                torch.save(
                    model.state_dict(),
                    os.path.join(
                        model_dir,
                        f"model-warmup-{ds_name}-{mod_name}-epoch{epoch}.pt",
                    ),
                )

                torch.save(
                    optimizer.state_dict(),
                    os.path.join(
                        model_dir,
                        f"opt-warmup-{ds_name}-{mod_name}-epoch{epoch}.tar",
                    ),
                )

        # Report full training time

        end_tot = time.time()

        total_time = end_tot - start_tot

        f.write("Total training time: " + str(total_time) + "\n")


def adjust_learning_rate(
    optimizer: Optimizer, epoch: int, args: NamedTuple
) -> None:
    """

    Sets the learning rate of the optimizer based on the current epoch


    :param optimizer: optimizer with learning rate being set

    :param epoch: current epoch

    :param args: program arguments
    """
    lr = args.lr

    if args.lr_schedule == "decay":
        if epoch >= 50:
            lr = args.lr * 0.1

        if epoch >= 75:
            lr = args.lr * 0.01

    elif args.lr_schedule == "scheduled":
        if epoch >= 24:
            lr = args.lr * 0.1

        if epoch >= 26:
            lr = args.lr

        if epoch >= 64:
            lr = args.lr * 0.1

        if epoch >= 66:
            lr = args.lr

        if epoch >= 104:
            lr = args.lr * 0.1

        if epoch >= 106:
            lr = args.lr

        if epoch >= 137:
            lr = args.lr * 0.1

        if epoch >= 139:
            lr = args.lr * 0.01

    elif args.lr_schedule == "cosine":
        lr = 0.2

        lr = lr * 0.5 * (1 + np.cos((epoch - 1) / args.epochs * np.pi))

    elif args.lr_schedule == "cyclic-10":
        if epoch % 10 < 2 or epoch % 10 == 9:
            lr = args.lr

        elif epoch % 10 < 4 or epoch % 10 > 6:
            lr = args.lr * 0.1

        elif epoch % 10 == 5:
            lr = args.lr * 0.001

        else:
            lr = args.lr * 0.01

    elif args.lr_schedule == "cyclic-5":
        if epoch % 5 == 0:
            lr = args.lr

        elif epoch % 5 == 1 or epoch % 5 == 4:
            lr = args.lr * 0.1

        else:
            lr = args.lr * 0.01

    for param_group in optimizer.param_groups:
        param_group["lr"] = lr



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from scipy import stats as st

from ..attack import Attack


class TIFGSM(Attack):
    r"""
    TIFGSM in the paper 'Evading Defenses to Transferable Adversarial Examples by Translation-Invariant Attacks'
    [https://arxiv.org/abs/1904.02884]

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)
        alpha (float): step size. (Default: 2/255)
        steps (int): number of iterations. (Default: 10)
        decay (float): momentum factor. (Default: 0.0)
        kernel_name (str): kernel name. (Default: gaussian)
        len_kernel (int): kernel length.  (Default: 15, which is the best according to the paper)
        nsig (int): radius of gaussian kernel. (Default: 3; see Section 3.2.2 in the paper for explanation)
        resize_rate (float): resize factor used in input diversity. (Default: 0.9)
        diversity_prob (float) : the probability of applying input diversity. (Default: 0.5)
        random_start (bool): using random initialization of delta. (Default: False)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`, `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.TIFGSM(model, eps=8/255, alpha=2/255, steps=10, decay=1.0, resize_rate=0.9, diversity_prob=0.7, random_start=False)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        eps=8 / 255,
        alpha=2 / 255,
        steps=10,
        decay=0.0,
        kernel_name="gaussian",
        len_kernel=15,
        nsig=3,
        resize_rate=0.9,
        diversity_prob=0.5,
        random_start=False,
    ):
        super().__init__("TIFGSM", model)
        self.eps = eps
        self.steps = steps
        self.decay = decay
        self.alpha = alpha
        self.resize_rate = resize_rate
        self.diversity_prob = diversity_prob
        self.random_start = random_start
        self.kernel_name = kernel_name
        self.len_kernel = len_kernel
        self.nsig = nsig
        self.stacked_kernel = torch.from_numpy(self.kernel_generation())
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        loss = nn.CrossEntropyLoss()
        momentum = torch.zeros_like(images).detach().to(self.device)
        stacked_kernel = self.stacked_kernel.to(self.device)

        adv_images = images.clone().detach()

        if self.random_start:
            # Starting at a uniformly random point
            adv_images = adv_images + torch.empty_like(adv_images).uniform_(
                -self.eps, self.eps
            )
            adv_images = torch.clamp(adv_images, min=0, max=1).detach()

        for _ in range(self.steps):
            adv_images.requires_grad = True
            outputs = self.get_logits(self.input_diversity(adv_images))

            # Calculate loss
            if self.targeted:
                cost = -loss(outputs, target_labels)
            else:
                cost = loss(outputs, labels)

            # Update adversarial images
            grad = torch.autograd.grad(
                cost, adv_images, retain_graph=False, create_graph=False
            )[0]
            # depth wise conv2d
            grad = F.conv2d(
                grad, stacked_kernel, stride=1, padding="same", groups=3
            )
            grad = grad / torch.mean(
                torch.abs(grad), dim=(1, 2, 3), keepdim=True
            )
            grad = grad + momentum * self.decay
            momentum = grad

            adv_images = adv_images.detach() + self.alpha * grad.sign()
            delta = torch.clamp(
                adv_images - images, min=-self.eps, max=self.eps
            )
            adv_images = torch.clamp(images + delta, min=0, max=1).detach()

        return adv_images

    def kernel_generation(self):
        if self.kernel_name == "gaussian":
            kernel = self.gkern(self.len_kernel, self.nsig).astype(np.float32)
        elif self.kernel_name == "linear":
            kernel = self.lkern(self.len_kernel).astype(np.float32)
        elif self.kernel_name == "uniform":
            kernel = self.ukern(self.len_kernel).astype(np.float32)
        else:
            raise NotImplementedError

        stack_kernel = np.stack([kernel, kernel, kernel])
        stack_kernel = np.expand_dims(stack_kernel, 1)
        return stack_kernel

    def gkern(self, kernlen=15, nsig=3):
        """Returns a 2D Gaussian kernel array."""
        x = np.linspace(-nsig, nsig, kernlen)
        kern1d = st.norm.pdf(x)
        kernel_raw = np.outer(kern1d, kern1d)
        kernel = kernel_raw / kernel_raw.sum()
        return kernel

    def ukern(self, kernlen=15):
        kernel = np.ones((kernlen, kernlen)) * 1.0 / (kernlen * kernlen)
        return kernel

    def lkern(self, kernlen=15):
        kern1d = 1 - np.abs(
            np.linspace((-kernlen + 1) / 2, (kernlen - 1) / 2, kernlen)
            / (kernlen + 1)
            * 2
        )
        kernel_raw = np.outer(kern1d, kern1d)
        kernel = kernel_raw / kernel_raw.sum()
        return kernel

    def input_diversity(self, x):
        img_size = x.shape[-1]
        img_resize = int(img_size * self.resize_rate)

        if self.resize_rate < 1:
            img_size = img_resize
            img_resize = x.shape[-1]

        rnd = torch.randint(
            low=img_size, high=img_resize, size=(1,), dtype=torch.int32
        )
        rescaled = F.interpolate(
            x, size=[rnd, rnd], mode="bilinear", align_corners=False
        )
        h_rem = img_resize - rnd
        w_rem = img_resize - rnd
        pad_top = torch.randint(
            low=0, high=h_rem.item(), size=(1,), dtype=torch.int32
        )
        pad_bottom = h_rem - pad_top
        pad_left = torch.randint(
            low=0, high=w_rem.item(), size=(1,), dtype=torch.int32
        )
        pad_right = w_rem - pad_left

        padded = F.pad(
            rescaled,
            [
                pad_left.item(),
                pad_right.item(),
                pad_top.item(),
                pad_bottom.item(),
            ],
            value=0,
        )

        return padded if torch.rand(1) < self.diversity_prob else x



import os
import sys

father_dir = os.path.join(
    "/", *os.path.realpath(__file__).split(os.path.sep)[:-2]
)
# print(father_dir)
if father_dir not in sys.path:
    sys.path.append(father_dir)
from collections import OrderedDict

import torch
from tqdm import tqdm

from airtk.defense.yopo.utils_misc import (
    AvgMeter,
    torch_accuracy,
)


def train_one_epoch(
    net,
    batch_generator,
    optimizer,
    criterion,
    DEVICE=torch.device("cuda:0"),
    descrip_str="Training",
    AttackMethod=None,
    adv_coef=1.0,
):
    """

    :param attack_freq:  Frequencies of training with adversarial examples. -1 indicates natural training
    :param AttackMethod: the attack method, None represents natural training
    :return:  None    #(clean_acc, adv_acc)
    """
    net.train()
    pbar = tqdm(batch_generator)
    advacc = -1
    advloss = -1
    cleanacc = -1
    cleanloss = -1
    pbar.set_description(descrip_str)
    for i, (data, label) in enumerate(pbar):
        data = data.to(DEVICE)
        label = label.to(DEVICE)

        optimizer.zero_grad()

        pbar_dic = OrderedDict()
        TotalLoss = 0

        if AttackMethod is not None:
            adv_inp = AttackMethod.attack(net, data, label)
            optimizer.zero_grad()
            net.train()
            pred = net(adv_inp)
            loss = criterion(pred, label)

            acc = torch_accuracy(pred, label, (1,))
            advacc = acc[0].item()
            advloss = loss.item()
            # TotalLoss = TotalLoss + loss * adv_coef
            (loss * adv_coef).backward()

        pred = net(data)

        loss = criterion(pred, label)
        # TotalLoss = TotalLoss + loss
        loss.backward()
        # TotalLoss.backward()
        # param = next(net.parameters())
        # grad_mean = torch.mean(param.grad)

        optimizer.step()
        acc = torch_accuracy(pred, label, (1,))
        cleanacc = acc[0].item()
        cleanloss = loss.item()
        # pbar_dic['grad'] = '{}'.format(grad_mean)
        pbar_dic["Acc"] = f"{cleanacc:.2f}"
        pbar_dic["loss"] = f"{cleanloss:.2f}"
        pbar_dic["AdvAcc"] = f"{advacc:.2f}"
        pbar_dic["Advloss"] = f"{advloss:.2f}"
        pbar.set_postfix(pbar_dic)


def eval_one_epoch(
    net, batch_generator, DEVICE=torch.device("cuda:0"), AttackMethod=None
):
    net.eval()
    pbar = tqdm(batch_generator)
    clean_accuracy = AvgMeter()
    adv_accuracy = AvgMeter()

    pbar.set_description("Evaluating")
    for data, label in pbar:
        data = data.to(DEVICE)
        label = label.to(DEVICE)

        with torch.no_grad():
            pred = net(data)
            acc = torch_accuracy(pred, label, (1,))
            clean_accuracy.update(acc[0].item())

        if AttackMethod is not None:
            adv_inp = AttackMethod.attack(net, data, label)

            with torch.no_grad():
                pred = net(adv_inp)
                acc = torch_accuracy(pred, label, (1,))
                adv_accuracy.update(acc[0].item())

        pbar_dic = OrderedDict()
        pbar_dic["CleanAcc"] = f"{clean_accuracy.mean:.2f}"
        pbar_dic["AdvAcc"] = f"{adv_accuracy.mean:.2f}"

        pbar.set_postfix(pbar_dic)

        adv_acc = adv_accuracy.mean if AttackMethod is not None else 0
    return clean_accuracy.mean, adv_acc



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).

import copy
import os
import time
from collections import namedtuple

import torch
from autoattack import AutoAttack
from torch import nn, optim
from warmup_round import main_warmup

from airtk.attack import create_attack
from airtk.helper_functions import (
    adjust_learning_rate,
    cw_whitebox,
    eval_clean,
    get_model,
    load_data,
    mim_whitebox,
    robust_eval,
)
from airtk.loss import accuracy, standard_loss

"""Trains a model using the Various Attacks method.

Models: ResNet-18, ResNet-34, ResNet-50, ResNet-101, ResNet-152, WideResNet-34

Datasets: CIFAR-10, CIFAR-100"""


class VaTraining:
    def __init__(
        self,
        ds_name: str,
        model_name: str,
        batch_size: int = 128,
        epochs: int = 200,
        weight_decay: float = 2e-4,
        lr: float = 0.1,
        momentum: float = 0.9,
        seed: int = 1,
        log_interval: int = 1,
        model_dir: str = "./data/model",
        save_freq: int = 1,
        swa_warmup: int = 50,
        swa: bool = True,
        lr_schedule: str = "decay",
    ) -> None:
        ArgsPrototype = namedtuple(
            "ArgsPrototype",
            [
                "batch_size",
                "test_batch_size",
                "epochs",
                "weight_decay",
                "lr",
                "momentum",
                "no_cuda",
                "seed",
                "log_interval",
                "model_dir",
                "save_freq",
                "swa_warmup",
                "swa",
                "lr_schedule",
            ],
        )
        self._args = ArgsPrototype(
            batch_size,
            batch_size,
            epochs,
            weight_decay,
            lr,
            momentum,
            False,
            seed,
            log_interval,
            model_dir,
            save_freq,
            swa_warmup,
            swa,
            lr_schedule,
        )
        self._ds_name = ds_name
        self._model_name = model_name

    def __call__(self) -> None:
        main(self._ds_name, self._model_name, 0, self._args)


# Define settings


def train(
    args, model, device, train_loader, optimizer, epoch, attacks, ds_name
):
    """
    Trains one epoch by calculating the loss for each batch based on the given set
    of attacks and updating the model

    :param args: set of arguments including learning rate, log interval, etc
    :param model: model to train
    :param device: the current device
    :param train_loader: data loader containing the training dataset
    :param optimizer: optimizer to train
    :param epoch: current epoch of training
    :param attacks: dictionary of attack types based on class
    :param ds_name: name of dataset
    """

    model.train()  # Set the model to training mode

    # For each batch of images in the train loader
    for batch_idx, sample in enumerate(train_loader):
        # Extract the image and classification details
        if ds_name == "cifar100":
            data, target, coarse = sample
        else:
            data, target = sample

        # Set the data/target to the current device
        data, target = (
            data.to(device, non_blocking=True),
            target.to(device, non_blocking=True),
        )

        # Determine the 'class' of the batch by looking at the most common class
        if ds_name == "cifar100":
            avg_cls = int(torch.mode(coarse)[0])
        else:
            avg_cls = int(torch.mode(target)[0])

        # Determine the attack based on the most common class
        attack = attacks[avg_cls]

        optimizer.zero_grad()  # zero optimizer gradients

        # Calculate loss using the specified attack
        loss, batch_metrics = va_loss(
            model, data, target, optimizer, attack, device, ds_name
        )

        # Update model parameters
        loss.backward()
        optimizer.step()

        # Print training progress
        if batch_idx % args.log_interval == 0:
            print(
                f"Train Epoch: {epoch} [{batch_idx * len(data)}/{len(train_loader.dataset)} ({100.0 * batch_idx / len(train_loader):.0f}%)]\tLoss: {loss.item():.6f}"
            )


def va_loss(model, x, y, optimizer, att, device, ds, keep_clean=False):
    """
    Adversarial training loss function for later various atts epoch, which simply
    calculates the loss based on the given attack.

    :param model: trained image classifier
    :param x: batch of clean images
    :param y: labels associated with images in x
    :param optimizer: optimizer to be updated
    :param att: attack to be used
    :param device: current device
    :param ds: name of dataset
    :param keep_clean: True if clean examples are kept, False default

    :return loss to the classifier
    :return metrics of the batch
    """

    # Set the criterion to cross-entropy loss
    criterion = nn.CrossEntropyLoss()

    # Create the adversary for AutoAttack
    adversary = AutoAttack(
        model, norm="Linf", eps=0.031, version="custom", verbose=False
    )

    # Generate adversarial example based on specified attack
    if att == "none":
        return standard_loss(model, x, y, optimizer)
    elif att == "linf-pgd-40":
        attack = create_attack(model, criterion, "linf-pgd", 0.03, 40, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "linf-pgd-20":
        attack = create_attack(model, criterion, "linf-pgd", 0.03, 20, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "linf-pgd-7":
        attack = create_attack(model, criterion, "linf-pgd", 0.03, 7, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "linf-pgd-3":
        attack = create_attack(model, criterion, "linf-pgd", 0.03, 3, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "fgsm-40":
        attack = create_attack(model, criterion, "fgsm", 0.03, 40, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "fgsm-20":
        attack = create_attack(model, criterion, "fgsm", 0.03, 20, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "fgsm-7":
        attack = create_attack(model, criterion, "fgsm", 0.03, 7, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "cw":
        x_adv = cw_whitebox(model, x, y, device, ds)
    elif att == "mim":
        x_adv = mim_whitebox(model, x, y, device)
    elif att == "l2-pgd-40":
        attack = create_attack(model, criterion, "l2-pgd", 0.03, 40, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "l2-pgd-20":
        attack = create_attack(model, criterion, "l2-pgd", 0.03, 20, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "l2-pgd-7":
        attack = create_attack(model, criterion, "l2-pgd", 0.03, 7, 0.01)
        x_adv, _ = attack.perturb(x, y)
    elif att == "autoattack":
        aa_attack = AutoAttack(
            model, norm="Linf", eps=0.031, version="standard", verbose=False
        )
        x_adv = aa_attack.run_standard_evaluation(x, y)
    elif att == "apgd-ce":
        adversary.attacks_to_run = ["apgd-ce"]
        x_adv = adversary.run_standard_evaluation(x, y)
    elif att == "apgd-dlr":
        adversary.attacks_to_run = ["apgd-dlr"]
        x_adv = adversary.run_standard_evaluation(x, y)
    elif att == "apgd-t":
        adversary.attacks_to_run = ["apgd-t"]
        x_adv = adversary.run_standard_evaluation(x, y)
    elif att == "fab":
        adversary.attacks_to_run = ["fab"]
        x_adv = adversary.run_standard_evaluation(x, y)
    elif att == "fab-t":
        adversary.attacks_to_run = ["fab-t"]
        x_adv = adversary.run_standard_evaluation(x, y)
    elif att == "square":
        adversary.attacks_to_run = ["square"]
        x_adv = adversary.run_standard_evaluation(x, y)
    else:
        print(att)
        raise NotImplementedError

    optimizer.zero_grad()  # zero out the gradients

    # Correctly format x_adv and y_adv
    if keep_clean:
        x_adv = torch.cat((x, x_adv), dim=0)
        y_adv = torch.cat((y, y), dim=0)
    else:
        y_adv = y

    # Calculates the adversarial output and CE loss
    out = model(x_adv)
    criterion = nn.CrossEntropyLoss()
    loss = criterion(out, y_adv)

    preds = out.detach()  # correctly format adversarial output

    # Calculate batch metrics to be returned
    batch_metrics = {"loss": loss.item()}
    if keep_clean:
        preds_clean, preds_adv = preds[: len(x)], preds[len(x) :]
        batch_metrics.update(
            {
                "clean_acc": accuracy(y, preds_clean),
                "adversarial_acc": accuracy(y, preds_adv),
            }
        )
    else:
        batch_metrics.update({"adversarial_acc": accuracy(y, preds)})

    return loss, batch_metrics


def epochs_define_attacks(epoch, dataset):
    """
    Defines a dictionary of attacks based on the current epoch.

    :param epoch: current epoch
    :param dataset: name of dataset

    :return attacks: dictionary of attacks
    """

    attacks = {}
    if dataset == "cifar10":
        if epoch <= 5:
            for i in range(4):
                attacks[i] = "l2-pgd-7"
            for i in range(4, 7):
                attacks[i] = "l2-pgd-40"
            for i in range(7, 10):
                attacks[i] = "l2-pgd-20"
        elif epoch <= 10:
            for i in range(4):
                attacks[i] = "l2-pgd-20"
            for i in range(4, 7):
                attacks[i] = "l2-pgd-7"
            for i in range(7, 10):
                attacks[i] = "l2-pgd-40"
        elif epoch <= 15:
            for i in range(4):
                attacks[i] = "l2-pgd-40"
            for i in range(4, 7):
                attacks[i] = "l2-pgd-20"
            for i in range(7, 10):
                attacks[i] = "l2-pgd-7"
        elif epoch <= 25:
            for i in range(10):
                attacks[i] = "linf-pgd-3"
        elif epoch <= 30:
            for i in range(3):
                attacks[i] = "linf-pgd-7"
            for i in range(3, 6):
                attacks[i] = "linf-pgd-20"
            for i in range(6, 8):
                attacks[i] = "linf-pgd-40"
            for i in range(8, 10):
                attacks[i] = "mim"
        elif epoch <= 35:
            for i in range(3):
                attacks[i] = "linf-pgd-20"
            for i in range(3, 6):
                attacks[i] = "linf-pgd-7"
            for i in range(6, 8):
                attacks[i] = "mim"
            for i in range(8, 10):
                attacks[i] = "linf-pgd-40"
        elif epoch <= 40:
            for i in range(3):
                attacks[i] = "linf-pgd-40"
            for i in range(3, 6):
                attacks[i] = "mim"
            for i in range(6, 8):
                attacks[i] = "linf-pgd-7"
            for i in range(8, 10):
                attacks[i] = "linf-pgd-20"
        elif epoch <= 45:
            for i in range(3):
                attacks[i] = "mim"
            for i in range(3, 6):
                attacks[i] = "linf-pgd-40"
            for i in range(6, 8):
                attacks[i] = "linf-pgd-20"
            for i in range(8, 10):
                attacks[i] = "linf-pgd-7"
        elif epoch <= 53:
            for i in range(2):
                attacks[i] = "apgd-t"
            for i in range(2, 5):
                attacks[i] = "cw"
            for i in range(5, 8):
                attacks[i] = "apgd-dlr"
            for i in range(8, 10):
                attacks[i] = "apgd-ce"
        elif epoch <= 61:
            for i in range(2):
                attacks[i] = "apgd-ce"
            for i in range(2, 5):
                attacks[i] = "apgd-t"
            for i in range(5, 8):
                attacks[i] = "cw"
            for i in range(8, 10):
                attacks[i] = "apgd-dlr"
        elif epoch <= 68:
            for i in range(2):
                attacks[i] = "apgd-dlr"
            for i in range(2, 5):
                attacks[i] = "apgd-ce"
            for i in range(5, 8):
                attacks[i] = "apgd-t"
            for i in range(8, 10):
                attacks[i] = "cw"
        elif epoch <= 75:
            for i in range(2):
                attacks[i] = "cw"
            for i in range(2, 5):
                attacks[i] = "apgd-dlr"
            for i in range(5, 8):
                attacks[i] = "apgd-ce"
            for i in range(8, 10):
                attacks[i] = "apgd-t"
        else:
            for i in range(10):
                attacks[i] = "autoattack"
    elif dataset == "cifar100":
        if epoch <= 25:
            for i in range(20):
                attacks[i] = "linf-pgd-3"
        elif epoch <= 45:
            for i in range(10):
                attacks[i] = "linf-pgd-7"
            for i in range(10, 20):
                attacks[i] = "mim"
        elif epoch <= 65:
            for i in range(10):
                attacks[i] = "mim"
            for i in range(10, 20):
                attacks[i] = "linf-pgd-7"
        elif epoch <= 85:
            for i in range(10):
                attacks[i] = "linf-pgd-20"
            for i in range(10, 20):
                attacks[i] = "linf-pgd-40"
        elif epoch <= 105:
            for i in range(10):
                attacks[i] = "linf-pgd-40"
            for i in range(10, 20):
                attacks[i] = "linf-pgd-20"
        elif epoch <= 140:
            for i in range(20):
                attacks[i] = "cw"
        else:
            for i in range(20):
                attacks[i] = "autoattack"
    else:
        raise NotImplementedError

    return attacks


def main(ds_name, mod_name, clean_epochs, args):
    """
    Main training method, which establishes the model/dataset before conducting training and
    clean testing for each epoch. Then reports final training time and conducts robustness tests
    on the final model.

    :param ds_name: dataset (cifar10 or cifar100)
    :param mod_name: training model architecture
    :param clean_pochs: number of epochs in warmup round, default 0
    """

    model_dir = args.model_dir
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)
    use_cuda = not args.no_cuda and torch.cuda.is_available()
    torch.manual_seed(args.seed)
    device = torch.device("cuda" if use_cuda else "cpu")
    kwargs = {"num_workers": 1, "pin_memory": True} if use_cuda else {}

    # Create file to print training progress
    filename = f"va-epochs-{ds_name}-{mod_name}-output.txt"
    f = open(filename, "a")

    # Initialize the model based on the specified architecture
    start_tot = time.time()  # start recording training time
    base_model = get_model(mod_name, ds_name, device)

    # Clean training/warmup round, if specified
    if clean_epochs == 0:
        model = copy.deepcopy(base_model).to(device)
    else:
        main_warmup(ds_name, mod_name, clean_epochs)
        base_model.load_state_dict(
            torch.load(
                os.path.join(
                    model_dir,
                    f"model-warmup-{ds_name}-{mod_name}-epoch{clean_epochs}.pt",
                )
            )
        )
        model = copy.deepcopy(base_model).to(device)

    # Set up the optimizer with the arguments
    optimizer = optim.SGD(
        model.parameters(),
        lr=args.lr,
        momentum=args.momentum,
        weight_decay=args.weight_decay,
    )

    # Set up the dataloaders
    train_loader, test_loader = load_data(ds_name, args, kwargs, coarse=True)

    # Save the untrained model and optimizer
    torch.save(
        model.state_dict(),
        os.path.join(
            model_dir,
            f"model-va-epochs-{ds_name}-{mod_name}-start.pt",
        ),
    )
    torch.save(
        optimizer.state_dict(),
        os.path.join(
            model_dir,
            f"opt-va-epochs-{ds_name}-{mod_name}-start.tar",
        ),
    )

    # Begin model training
    for epoch in range(1, args.epochs + 1):
        # Adjust the SGD learning rate based on the epoch
        adjust_learning_rate(optimizer, epoch, args)

        # Define the dictionary of attack types for the current epoch
        attacks = epochs_define_attacks(epoch, ds_name)

        # Calculate the time elapsed
        start = time.time()  # start recording training time
        train(
            args,
            model,
            device,
            train_loader,
            optimizer,
            epoch,
            attacks,
            ds_name,
        )
        end = time.time()  # stop recording training time
        epoch_time = end - start

        # Evaluation on natural and adversarial examples
        print(
            "================================================================"
        )
        f.write("Epoch " + str(epoch) + "\n")
        f.write("Current Attack Dictionary: ")
        f.write(str(attacks) + "\n")
        print("Time for Training: " + str(epoch_time))
        f.write("Time for Training: " + str(epoch_time) + "\n")
        eval_clean(model, device, train_loader, "train", ds_name, f)
        eval_clean(model, device, test_loader, "test", ds_name, f)
        robust_eval(model, device, test_loader, ds_name, f)
        print(
            "================================================================"
        )

        # Save the model (if designated)
        if epoch % args.save_freq == 0:
            torch.save(
                model.state_dict(),
                os.path.join(
                    model_dir,
                    f"model-va-epochs-{ds_name}-{mod_name}-epoch{epoch}.pt",
                ),
            )
            torch.save(
                optimizer.state_dict(),
                os.path.join(
                    model_dir,
                    f"opt-va-epochs-{ds_name}-{mod_name}-epoch{epoch}.tar",
                ),
            )

    # Report full training time
    end_tot = time.time()
    total_time = end_tot - start_tot
    print("Total training time: " + str(total_time))
    f.write("Total training time: " + str(total_time) + "\n")

    # Load the final trained model, make a deepcopy, and evaluate against FGSM and PGD attacks
    trained_model = get_model(mod_name, ds_name, device)

    trained_model.load_state_dict(
        torch.load(
            f"data/model/model-va-epochs-{ds_name}-{mod_name}-epoch{args.epochs}.pt"
        )
    )
    model_copy = copy.deepcopy(trained_model)
    robust_eval(model_copy, device, test_loader, ds_name, f)

    f.close()  # close output file



import torch
import torch.nn as nn
import torch.optim as optim

from ..attack import Attack


class CW(Attack):
    r"""
    CW in the paper 'Towards Evaluating the Robustness of Neural Networks'
    [https://arxiv.org/abs/1608.04644]

    Distance Measure : L2

    Arguments:
        model (nn.Module): model to attack.
        c (float): c in the paper. parameter for box-constraint. (Default: 1)    
            :math:`minimize \Vert\frac{1}{2}(tanh(w)+1)-x\Vert^2_2+c\cdot f(\frac{1}{2}(tanh(w)+1))`
        kappa (float): kappa (also written as 'confidence') in the paper. (Default: 0)
            :math:`f(x')=max(max\{Z(x')_i:i\neq t\} -Z(x')_t, - \kappa)`
        steps (int): number of steps. (Default: 50)
        lr (float): learning rate of the Adam optimizer. (Default: 0.01)

    .. warning:: With default c, you can't easily get adversarial images. Set higher c like 1.

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.CW(model, c=1, kappa=0, steps=50, lr=0.01)
        >>> adv_images = attack(images, labels)

    .. note:: Binary search for c is NOT IMPLEMENTED methods in the paper due to time consuming.

    """

    def __init__(self, model, c=1, kappa=0, steps=50, lr=0.01):
        super().__init__("CW", model)
        self.c = c
        self.kappa = kappa
        self.steps = steps
        self.lr = lr
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        # w = torch.zeros_like(images).detach() # Requires 2x times
        w = self.inverse_tanh_space(images).detach()
        w.requires_grad = True

        best_adv_images = images.clone().detach()
        best_L2 = 1e10 * torch.ones((len(images))).to(self.device)
        prev_cost = 1e10
        dim = len(images.shape)

        MSELoss = nn.MSELoss(reduction="none")
        Flatten = nn.Flatten()

        optimizer = optim.Adam([w], lr=self.lr)

        for step in range(self.steps):
            # Get adversarial images
            adv_images = self.tanh_space(w)

            # Calculate loss
            current_L2 = MSELoss(Flatten(adv_images), Flatten(images)).sum(dim=1)
            L2_loss = current_L2.sum()

            outputs = self.get_logits(adv_images)
            if self.targeted:
                f_loss = self.f(outputs, target_labels).sum()
            else:
                f_loss = self.f(outputs, labels).sum()

            cost = L2_loss + self.c * f_loss

            optimizer.zero_grad()
            cost.backward()
            optimizer.step()

            # Update adversarial images
            pre = torch.argmax(outputs.detach(), 1)
            if self.targeted:
                # We want to let pre == target_labels in a targeted attack
                condition = (pre == target_labels).float()
            else:
                # If the attack is not targeted we simply make these two values unequal
                condition = (pre != labels).float()

            # Filter out images that get either correct predictions or non-decreasing loss,
            # i.e., only images that are both misclassified and loss-decreasing are left
            mask = condition * (best_L2 > current_L2.detach())
            best_L2 = mask * current_L2.detach() + (1 - mask) * best_L2

            mask = mask.view([-1] + [1] * (dim - 1))
            best_adv_images = mask * adv_images.detach() + (1 - mask) * best_adv_images

            # Early stop when loss does not converge.
            # max(.,1) To prevent MODULO BY ZERO error in the next step.
            if step % max(self.steps // 10, 1) == 0:
                if cost.item() > prev_cost:
                    return best_adv_images
                prev_cost = cost.item()

        return best_adv_images

    def tanh_space(self, x):
        return 1 / 2 * (torch.tanh(x) + 1)

    def inverse_tanh_space(self, x):
        # torch.atanh is only for torch >= 1.7.0
        # atanh is defined in the range -1 to 1
        return self.atanh(torch.clamp(x * 2 - 1, min=-1, max=1))

    def atanh(self, x):
        return 0.5 * torch.log((1 + x) / (1 - x))

    # f-function in the paper
    def f(self, outputs, labels):
        one_hot_labels = torch.eye(outputs.shape[1]).to(self.device)[labels]

        # find the max logit other than the target class
        other = torch.max((1 - one_hot_labels) * outputs, dim=1)[0]
        # get the target class's logit
        real = torch.max(one_hot_labels * outputs, dim=1)[0]

        if self.targeted:
            return torch.clamp((other - real), min=-self.kappa)
        else:
            return torch.clamp((real - other), min=-self.kappa)



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
import argparse
import logging
import os
import time

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from airtk.defense.atawp.at_awp_utils_awp import (
    AdvWeightPerturb,
)
from airtk.model import ResNet18, WideResNet
from torch.autograd import Variable
from torchvision import datasets, transforms

CIFAR100_MEAN = (0.5070751592371323, 0.48654887331495095, 0.4409178433670343)
CIFAR100_STD = (0.2673342858792401, 0.2564384629170883, 0.27615047132568404)

mu = torch.tensor(CIFAR100_MEAN).view(3, 1, 1).cuda()
std = torch.tensor(CIFAR100_STD).view(3, 1, 1).cuda()


def normalize(X):
    return (X - mu) / std


upper_limit, lower_limit = 1, 0


def clamp(X, lower_limit, upper_limit):
    return torch.max(torch.min(X, upper_limit), lower_limit)


def mixup_data(x, y, alpha=1.0):
    """Returns mixed inputs, pairs of targets, and lambda"""
    if alpha > 0:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1

    batch_size = x.size()[0]
    index = torch.randperm(batch_size).cuda()

    mixed_x = lam * x + (1 - lam) * x[index, :]
    y_a, y_b = y, y[index]
    return mixed_x, y_a, y_b, lam


def mixup_criterion(criterion, pred, y_a, y_b, lam):
    return lam * criterion(pred, y_a) + (1 - lam) * criterion(pred, y_b)


def attack_pgd(
    model,
    X,
    y,
    epsilon,
    alpha,
    attack_iters,
    restarts,
    norm,
    early_stop=False,
    mixup=False,
    y_a=None,
    y_b=None,
    lam=None,
):
    max_loss = torch.zeros(y.shape[0]).cuda()
    max_delta = torch.zeros_like(X).cuda()
    for _ in range(restarts):
        delta = torch.zeros_like(X).cuda()
        if norm == "l_inf":
            delta.uniform_(-epsilon, epsilon)
        elif norm == "l_2":
            delta.normal_()
            d_flat = delta.view(delta.size(0), -1)
            n = d_flat.norm(p=2, dim=1).view(delta.size(0), 1, 1, 1)
            r = torch.zeros_like(n).uniform_(0, 1)
            delta *= r / n * epsilon
        else:
            raise ValueError
        delta = clamp(delta, lower_limit - X, upper_limit - X)
        delta.requires_grad = True
        for _ in range(attack_iters):
            output = model(normalize(X + delta))
            if early_stop:
                index = torch.where(output.max(1)[1] == y)[0]
            else:
                index = slice(None, None, None)
            if not isinstance(index, slice) and len(index) == 0:
                break
            if mixup:
                criterion = nn.CrossEntropyLoss()
                loss = mixup_criterion(
                    criterion, model(normalize(X + delta)), y_a, y_b, lam
                )
            else:
                loss = F.cross_entropy(output, y)
            loss.backward()
            grad = delta.grad.detach()
            d = delta[index, :, :, :]
            g = grad[index, :, :, :]
            x = X[index, :, :, :]
            if norm == "l_inf":
                d = torch.clamp(
                    d + alpha * torch.sign(g), min=-epsilon, max=epsilon
                )
            elif norm == "l_2":
                g_norm = torch.norm(g.view(g.shape[0], -1), dim=1).view(
                    -1, 1, 1, 1
                )
                scaled_g = g / (g_norm + 1e-10)
                d = (
                    (d + scaled_g * alpha)
                    .view(d.size(0), -1)
                    .renorm(p=2, dim=0, maxnorm=epsilon)
                    .view_as(d)
                )
            d = clamp(d, lower_limit - x, upper_limit - x)
            delta.data[index, :, :, :] = d
            delta.grad.zero_()
        if mixup:
            criterion = nn.CrossEntropyLoss(reduction="none")
            all_loss = mixup_criterion(
                criterion, model(normalize(X + delta)), y_a, y_b, lam
            )
        else:
            all_loss = F.cross_entropy(
                model(normalize(X + delta)), y, reduction="none"
            )
        max_delta[all_loss >= max_loss] = delta.detach()[all_loss >= max_loss]
        max_loss = torch.max(max_loss, all_loss)
    return max_delta


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--model", default="PreActResNet18")
    parser.add_argument("--l2", default=0, type=float)
    parser.add_argument("--l1", default=0, type=float)
    parser.add_argument("--batch-size", default=128, type=int)
    parser.add_argument("--data-dir", default="../cifar100-data", type=str)
    parser.add_argument("--epochs", default=200, type=int)
    parser.add_argument(
        "--lr-schedule",
        default="piecewise",
        choices=["superconverge", "piecewise"],
    )
    parser.add_argument("--lr-max", default=0.1, type=float)
    parser.add_argument(
        "--attack", default="pgd", type=str, choices=["pgd", "none"]
    )
    parser.add_argument("--epsilon", default=8, type=int)
    parser.add_argument("--attack-iters", default=10, type=int)
    parser.add_argument("--attack-iters-test", default=20, type=int)
    parser.add_argument("--restarts", default=1, type=int)
    parser.add_argument("--pgd-alpha", default=2, type=float)
    parser.add_argument(
        "--norm", default="l_inf", type=str, choices=["l_inf", "l_2"]
    )
    parser.add_argument("--fname", default="cifar100_model", type=str)
    parser.add_argument("--seed", default=0, type=int)
    parser.add_argument("--half", action="store_true")
    parser.add_argument("--width-factor", default=10, type=int)
    parser.add_argument("--resume", default=0, type=int)
    parser.add_argument("--cutout", action="store_true")
    parser.add_argument("--cutout-len", type=int)
    parser.add_argument("--mixup", action="store_true")
    parser.add_argument("--mixup-alpha", type=float)
    parser.add_argument("--chkpt-iters", default=10, type=int)
    parser.add_argument("--awp-gamma", default=0.01, type=float)
    parser.add_argument("--awp-warmup", default=0, type=int)
    return parser.parse_args()


def main():
    args = get_args()
    if args.awp_gamma <= 0.0:
        args.awp_warmup = np.infty

    if not os.path.exists(args.fname):
        os.makedirs(args.fname)

    logger = logging.getLogger(__name__)
    logging.basicConfig(
        format="[%(asctime)s] - %(message)s",
        datefmt="%Y/%m/%d %H:%M:%S",
        level=logging.DEBUG,
        handlers=[
            logging.FileHandler(os.path.join(args.fname, "output.log")),
            logging.StreamHandler(),
        ],
    )

    logger.info(args)

    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    torch.cuda.manual_seed(args.seed)

    train_transform = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.RandomRotation(15),
            transforms.ToTensor(),
        ]
    )
    test_transform = transforms.Compose(
        [
            transforms.ToTensor(),
        ]
    )
    num_workers = 2
    train_dataset = datasets.CIFAR100(
        args.data_dir, train=True, transform=train_transform, download=True
    )
    test_dataset = datasets.CIFAR100(
        args.data_dir, train=False, transform=test_transform, download=True
    )
    train_loader = torch.utils.data.DataLoader(
        dataset=train_dataset,
        batch_size=args.batch_size,
        shuffle=True,
        pin_memory=True,
        num_workers=num_workers,
    )
    test_loader = torch.utils.data.DataLoader(
        dataset=test_dataset,
        batch_size=args.batch_size,
        shuffle=False,
        pin_memory=True,
        num_workers=2,
    )

    epsilon = args.epsilon / 255.0
    pgd_alpha = args.pgd_alpha / 255.0

    if args.model == "PreActResNet18":
        model = ResNet18(num_classes=100)
        proxy = ResNet18(num_classes=100)
    elif args.model == "WideResNet":
        model = WideResNet(
            34, 100, widen_factor=args.width_factor, dropRate=0.0
        )
        proxy = WideResNet(
            34, 100, widen_factor=args.width_factor, dropRate=0.0
        )
    else:
        raise ValueError("Unknown model")

    model = model.cuda()
    proxy = proxy.cuda()

    if args.l2:
        decay, no_decay = [], []
        for name, param in model.named_parameters():
            if "bn" not in name and "bias" not in name:
                decay.append(param)
            else:
                no_decay.append(param)
        params = [
            {"params": decay, "weight_decay": args.l2},
            {"params": no_decay, "weight_decay": 0},
        ]
    else:
        params = model.parameters()

    opt = torch.optim.SGD(
        params, lr=args.lr_max, momentum=0.9, weight_decay=5e-4
    )
    proxy_opt = torch.optim.SGD(proxy.parameters(), lr=0.01)
    awp_adversary = AdvWeightPerturb(
        model=model, proxy=proxy, proxy_optim=proxy_opt, gamma=args.awp_gamma
    )

    criterion = nn.CrossEntropyLoss()

    epochs = args.epochs

    if args.lr_schedule == "superconverge":
        lr_schedule = lambda t: np.interp(
            [t], [0, args.epochs * 2 // 5, args.epochs], [0, args.lr_max, 0]
        )[0]
        # lr_schedule = lambda t: np.interp([t], [0, args.epochs], [0, args.lr_max])[0]
    elif args.lr_schedule == "piecewise":

        def lr_schedule(t):
            if t / args.epochs < 0.5:
                return args.lr_max
            elif t / args.epochs < 0.75:
                return args.lr_max / 10.0
            else:
                return args.lr_max / 100.0

    best_test_robust_acc = 0
    if args.resume:
        start_epoch = args.resume
        model.load_state_dict(
            torch.load(os.path.join(args.fname, f"model_{start_epoch-1}.pth"))
        )
        opt.load_state_dict(
            torch.load(os.path.join(args.fname, f"opt_{start_epoch-1}.pth"))
        )
        logger.info(f"Resuming at epoch {start_epoch}")

        if os.path.exists(os.path.join(args.fname, "model_best.pth")):
            best_test_robust_acc = torch.load(
                os.path.join(args.fname, "model_best.pth")
            )["test_robust_acc"]
    else:
        start_epoch = 0

    logger.info(
        "Epoch \t Train Time \t Test Time \t LR \t \t Train Loss \t Train Acc \t Train Robust Loss \t Train Robust Acc \t Test Loss \t Test Acc \t Test Robust Loss \t Test Robust Acc"
    )
    for epoch in range(start_epoch, epochs):
        model.train()
        start_time = time.time()
        train_loss = 0
        train_acc = 0
        train_robust_loss = 0
        train_robust_acc = 0
        train_n = 0
        for i, (X, y) in enumerate(train_loader):
            X, y = X.cuda(), y.cuda()
            if args.mixup:
                X, y_a, y_b, lam = mixup_data(X, y, args.mixup_alpha)
                X, y_a, y_b = map(Variable, (X, y_a, y_b))
            lr = lr_schedule(epoch + (i + 1) / len(train_loader))
            opt.param_groups[0].update(lr=lr)

            if args.attack == "pgd":
                # Random initialization
                if args.mixup:
                    delta = attack_pgd(
                        model,
                        X,
                        y,
                        epsilon,
                        pgd_alpha,
                        args.attack_iters,
                        args.restarts,
                        args.norm,
                        mixup=True,
                        y_a=y_a,
                        y_b=y_b,
                        lam=lam,
                    )
                else:
                    delta = attack_pgd(
                        model,
                        X,
                        y,
                        epsilon,
                        pgd_alpha,
                        args.attack_iters,
                        args.restarts,
                        args.norm,
                    )
                delta = delta.detach()
            # Standard training
            elif args.attack == "none":
                delta = torch.zeros_like(X)
            X_adv = normalize(
                torch.clamp(
                    X + delta[: X.size(0)], min=lower_limit, max=upper_limit
                )
            )

            model.train()
            # calculate adversarial weight perturbation and perturb it
            if epoch >= args.awp_warmup:
                # not compatible to mixup currently.
                assert not args.mixup
                awp = awp_adversary.calc_awp(inputs_adv=X_adv, targets=y)
                awp_adversary.perturb(awp)

            robust_output = model(X_adv)
            if args.mixup:
                robust_loss = mixup_criterion(
                    criterion, robust_output, y_a, y_b, lam
                )
            else:
                robust_loss = criterion(robust_output, y)

            if args.l1:
                for name, param in model.named_parameters():
                    if "bn" not in name and "bias" not in name:
                        robust_loss += args.l1 * param.abs().sum()

            opt.zero_grad()
            robust_loss.backward()
            opt.step()

            if epoch >= args.awp_warmup:
                awp_adversary.restore(awp)

            output = model(normalize(X))
            if args.mixup:
                loss = mixup_criterion(criterion, output, y_a, y_b, lam)
            else:
                loss = criterion(output, y)

            train_robust_loss += robust_loss.item() * y.size(0)
            train_robust_acc += (robust_output.max(1)[1] == y).sum().item()
            train_loss += loss.item() * y.size(0)
            train_acc += (output.max(1)[1] == y).sum().item()
            train_n += y.size(0)

        train_time = time.time()

        model.eval()
        test_loss = 0
        test_acc = 0
        test_robust_loss = 0
        test_robust_acc = 0
        test_n = 0
        for i, (X, y) in enumerate(test_loader):
            X, y = X.cuda(), y.cuda()

            # Random initialization
            if args.attack == "none":
                delta = torch.zeros_like(X)
            else:
                delta = attack_pgd(
                    model,
                    X,
                    y,
                    epsilon,
                    pgd_alpha,
                    args.attack_iters_test,
                    args.restarts,
                    args.norm,
                )
            delta = delta.detach()

            robust_output = model(
                normalize(
                    torch.clamp(
                        X + delta[: X.size(0)],
                        min=lower_limit,
                        max=upper_limit,
                    )
                )
            )
            robust_loss = criterion(robust_output, y)

            output = model(normalize(X))
            loss = criterion(output, y)

            test_robust_loss += robust_loss.item() * y.size(0)
            test_robust_acc += (robust_output.max(1)[1] == y).sum().item()
            test_loss += loss.item() * y.size(0)
            test_acc += (output.max(1)[1] == y).sum().item()
            test_n += y.size(0)

        test_time = time.time()
        logger.info(
            "%d \t %.1f \t \t %.1f \t \t %.4f \t %.4f \t %.4f \t %.4f \t \t %.4f \t \t %.4f \t %.4f \t %.4f \t \t %.4f",
            epoch,
            train_time - start_time,
            test_time - train_time,
            lr,
            train_loss / train_n,
            train_acc / train_n,
            train_robust_loss / train_n,
            train_robust_acc / train_n,
            test_loss / test_n,
            test_acc / test_n,
            test_robust_loss / test_n,
            test_robust_acc / test_n,
        )

        if (epoch + 1) % args.chkpt_iters == 0 or epoch + 1 == epochs:
            torch.save(
                model.state_dict(),
                os.path.join(args.fname, f"model_{epoch}.pth"),
            )
            torch.save(
                opt.state_dict(), os.path.join(args.fname, f"opt_{epoch}.pth")
            )

        # save best
        if test_robust_acc / test_n > best_test_robust_acc:
            torch.save(
                {
                    "state_dict": model.state_dict(),
                    "test_robust_acc": test_robust_acc / test_n,
                    "test_robust_loss": test_robust_loss / test_n,
                    "test_loss": test_loss / test_n,
                    "test_acc": test_acc / test_n,
                },
                os.path.join(args.fname, "model_best.pth"),
            )
            best_test_robust_acc = test_robust_acc / test_n
    return model


def main_atawp100():
    model = main()

    model_dir = "../data/model"
    torch.save(
        model.state_dict(),
        os.path.join(
            model_dir, "model-at-awp-{}-{}.pt".format("cifar100", "res18")
        ),
    )

    target_model = ResNet18(num_classes=100)
    from collections import OrderedDict

    checkpoint = torch.load(
        os.path.join(
            model_dir, "model-at-awp-{}-{}.pt".format("cifar100", "res18")
        )
    )
    try:
        target_model.load_state_dict(checkpoint)
    except Exception:
        new_state_dict = OrderedDict()
        for k, v in checkpoint.items():
            name = k[7:]  # remove `module.`
            new_state_dict[name] = v
        target_model.load_state_dict(new_state_dict, False)

    target_model = torch.nn.DataParallel(target_model).cuda()

    torch.save(
        target_model.module.state_dict(),
        os.path.join(
            model_dir, "model-at-awp-{}-{}.pt".format("cifar100", "res18")
        ),
    )


if __name__ == "__main__":
    main_atawp100()



from .lenet import LeNet, LeNetPP
from .mnist_ates import MNIST_ATES
from .mnist_dat import MNIST_DAT
from .mnist_fast import MNIST_Fast
from .preactresnet import PreActBlock, PreActResNet
from .resnet import ResBasicBlock, ResNet
from .densenet import DenseNet, Bottleneck
from .vgg import VGG
from .wideresnet import WideResNet


def load_model(model_name, n_classes):
    if model_name == "LeNet":
        return LeNet(n_classes)

    if model_name == "LeNetPP":
        return LeNetPP(n_classes)

    elif model_name == "MNIST_ATES":
        return MNIST_ATES(n_classes)

    elif model_name == "MNIST_DAT":
        return MNIST_DAT(n_classes)

    elif model_name == "MNIST_Fast":
        return MNIST_Fast(n_classes)

    elif model_name == "WRN28-10":
        model = WideResNet(
            depth=28, num_classes=n_classes, widen_factor=10, dropRate=0.0
        )

    elif model_name == "WRN34-10":
        model = WideResNet(
            depth=34, num_classes=n_classes, widen_factor=10, dropRate=0.0
        )

    elif model_name == "PRN18":
        model = PreActResNet(
            PreActBlock, num_blocks=[2, 2, 2, 2], num_classes=n_classes
        )

    elif model_name == "ResNet10":
        model = ResNet(ResBasicBlock, [1, 1, 1, 1], n_classes, in_channels=1)

    elif model_name == "ResNet18":
        model = ResNet(ResBasicBlock, [2, 2, 2, 2], n_classes)

    elif model_name == "ResNet34":
        model = ResNet(ResBasicBlock, [3, 4, 6, 3], n_classes)

    elif model_name == "ResNet50":
        model = ResNet(ResBasicBlock, [3, 4, 6, 3], n_classes)

    elif model_name == "ResNet101":
        model = ResNet(ResBasicBlock, [3, 4, 23, 3], n_classes)

    elif model_name == "ResNet152":
        model = ResNet(ResBasicBlock, [3, 8, 36, 3], n_classes)

    elif model_name == "DenseNet121":
        model = DenseNet(Bottleneck, [6, 12, 24, 16], growth_rate=32)

    elif model_name == "DenseNet169":
        model = DenseNet(Bottleneck, [6, 12, 32, 32], growth_rate=32)

    elif model_name == "DenseNet201":
        model = DenseNet(Bottleneck, [6, 12, 48, 32], growth_rate=32)

    elif model_name == "DenseNet161":
        model = DenseNet(Bottleneck, [6, 12, 36, 24], growth_rate=48)

    elif model_name == "VGG11":
        model = VGG("VGG11", n_classes)

    elif model_name == "VGG13":
        model = VGG("VGG13", n_classes)

    elif model_name == "VGG16":
        model = VGG("VGG16", n_classes)

    elif model_name == "VGG19":
        model = VGG("VGG19", n_classes)

    else:
        raise ValueError("Invalid model name.")

    print(model_name, "is loaded.")

    return model



from collections import OrderedDict

import torch
import torch.nn as nn

from ..attacks import FGSM, PGD, PGDL2, GN, AutoAttack, MultiAttack

from ..utils import get_accuracy
from .modules.normalize import Normalize


class RobModel(nn.Module):
    r"""
    Wrapper class for PyTorch models.
    """

    def __init__(
        self,
        model,
        n_classes,
        normalization_used={"mean": [0.0, 0.0, 0.0], "std": [1.0, 1.0, 1.0]},
        device=None,
    ):
        super().__init__()

        # Set device
        if device is None:
            device = next(model.parameters()).device

        # Set n_class
        assert isinstance(n_classes, int)
        self.register_buffer("n_classes", torch.tensor(n_classes))
        self.register_buffer("mean", torch.tensor(normalization_used["mean"]))
        self.register_buffer("std", torch.tensor(normalization_used["std"]))

        # Set model structure
        self.model = model.to(device)
        self.to(device)

    def add_normalize(self):
        if getattr(self, "mean") is not None and getattr(self, "std") is not None:
            pass
        else:
            raise ValueError("There is no normalization to be added.")

        self.model = nn.Sequential(
            Normalize(self.mean.cpu().numpy(), self.std.cpu().numpy()), self.model
        )

        delattr(self, "mean")
        delattr(self, "std")

    def forward(self, inputs, *args, **kargs):
        out = self.model(inputs, *args, **kargs)
        return out

    # Load from state dict
    def load_dict(self, save_path):
        state_dict = torch.load(save_path, map_location="cpu")
        self.load_state_dict_auto(state_dict["rmodel"])
        print("Model loaded.")

        if "record_info" in state_dict.keys():
            print("Record Info:")
            print(state_dict["record_info"])

    # DataParallel considered version of load_state_dict.
    def load_state_dict_auto(self, state_dict):
        state_dict = self._convert_dict_auto(state_dict)
        self.load_state_dict(state_dict)

    # Automatically changes pararell mode and non-parallel mode.
    def _convert_dict_auto(self, state_dict):
        keys = state_dict.keys()

        save_parallel = any(key.startswith("model.module.") for key in keys)
        curr_parallel = any(
            key.startswith("model.module.") for key in self.state_dict().keys()
        )
        if save_parallel and not curr_parallel:
            new_state_dict = {
                k.replace("model.module.", "model."): v for k, v in state_dict.items()
            }
            return new_state_dict
        elif curr_parallel and not save_parallel:
            new_state_dict = {
                k.replace("model.", "model.module."): v for k, v in state_dict.items()
            }
            return new_state_dict
        else:
            return state_dict

    def save_dict(self, save_path):
        save_dict = OrderedDict()
        save_dict["rmodel"] = self.state_dict()
        torch.save(save_dict, save_path)

    def set_parallel(self):
        self.model = torch.nn.DataParallel(self.model)
        return self

    def named_parameters_with_module(self):
        module_by_name = {}
        for name, module in self.named_modules():
            module_by_name[name] = module

        for name, param in self.named_parameters():
            if "." in name:
                module_name = name.rsplit(".", maxsplit=1)[0]
                yield name, param, module_by_name[module_name]
            else:
                yield name, param, None

    #################################################
    ############# Evaluate Robustness ###############
    #################################################
    @torch.no_grad()
    def eval_accuracy(self, data_loader):
        return get_accuracy(self, data_loader)

    def eval_rob_accuracy(self, data_loader, atk, **kargs):
        return atk.save(data_loader, return_verbose=True, **kargs)[0]

    def eval_rob_accuracy_gn(self, data_loader, std, **kargs):
        atk = GN(self, std=std)
        return self.eval_rob_accuracy(data_loader, atk, **kargs)

    def eval_rob_accuracy_fgsm(self, data_loader, eps, **kargs):
        atk = FGSM(self, eps=eps)
        return self.eval_rob_accuracy(data_loader, atk, **kargs)

    def eval_rob_accuracy_pgd(
        self,
        data_loader,
        eps,
        alpha,
        steps,
        random_start=True,
        restart_num=1,
        norm="Linf",
        **kargs
    ):
        if norm == "Linf":
            atk = PGD(
                self, eps=eps, alpha=alpha, steps=steps, random_start=random_start
            )
        elif norm == "L2":
            atk = PGDL2(
                self, eps=eps, alpha=alpha, steps=steps, random_start=random_start
            )
        else:
            raise ValueError("Invalid norm.")

        if restart_num > 1:
            atk = MultiAttack([atk] * restart_num)
        return self.eval_rob_accuracy(data_loader, atk, **kargs)

    def eval_rob_accuracy_autoattack(
        self, data_loader, eps, version="standard", norm="Linf", **kargs
    ):
        atk = AutoAttack(
            self, norm=norm, eps=eps, version=version, n_classes=self.n_classes
        )
        return self.eval_rob_accuracy(data_loader, atk, **kargs)

    ##############################################################
    ############# Evaluate Generalization Measures ###############
    ##############################################################
